import pytest

import ckan.plugins.toolkit as tk

from ckanext.files import config

from _pytest.monkeypatch import MonkeyPatch  # isort: skip # noqa: F401


class TestDefault:
    def test_default_value(self):
        """Verify that default value does not change accidentally.

        If you are modifying this test, don't forget to add changelog entry!

        """
        assert config.default_storage() == "default"

    @pytest.mark.ckan_config(config.DEFAULT_STORAGE, "test")
    def test_customized(self):
        """Default storage can be changed."""
        assert config.default_storage() == "test"


class TestStorages:
    def adapt_to_ckan_version(self, settings):
        # type: (dict[str, object]) -> dict[str, object]
        """CKAN v2.10 has max_size added to default storage by config
        declaration.

        """

        if tk.check_ckan_version("2.10"):
            settings.setdefault("max_size", 0)

        else:
            settings.pop("prefix")
            settings.pop("name")

        return settings

    def test_empty(self):
        """With no customization we have only storage defined by test.ini"""

        assert config.storages() == {
            "default": self.adapt_to_ckan_version(
                {
                    "type": "files:redis",
                    "prefix": "ckanext:files:test.ckan.net:file_content:",
                    "name": "default",
                },
            ),
        }

    def test_customized(self, monkeypatch, ckan_config):
        # type: (MonkeyPatch, dict[str, object]) -> None
        """Storage configuration grouped by the storage name."""
        patches = [
            ("default.type", "files:redis"),
            ("test.type", "test"),
            ("test.path", "somepath"),
            ("another.type", "fancy"),
            ("another.name_strategy", "hello"),
        ]  # type: list[tuple[str, str]]
        for key, value in patches:
            monkeypatch.setitem(ckan_config, config.STORAGE_PREFIX + key, value)

        storages = config.storages()

        assert storages == {
            "default": self.adapt_to_ckan_version(
                {
                    "type": "files:redis",
                    "prefix": "ckanext:files:test.ckan.net:file_content:",
                    "name": "default",
                },
            ),
            "test": {"type": "test", "path": "somepath"},
            "another": {"type": "fancy", "name_strategy": "hello"},
        }

    @pytest.mark.ckan_config(config.STORAGE_PREFIX + "another", "test")
    def test_non_setting(self):
        """Only `<PREFIX>.<NAME>.<OPTION>` settings are grouped. If `.<OPTION>`
        part is missing, we ignore this line.

        """

        storages = config.storages()

        assert storages == {
            "default": self.adapt_to_ckan_version(
                {
                    "type": "files:redis",
                    "prefix": "ckanext:files:test.ckan.net:file_content:",
                    "name": "default",
                },
            ),
        }



import sqlalchemy as sa
from werkzeug.utils import secure_filename

import ckan.plugins.toolkit as tk
from ckan import model
from ckan.logic import validate

from ckanext.files import exceptions, shared
from ckanext.files.base import Capability
from ckanext.files.model import File, Owner
from ckanext.files.utils import make_collector

from . import schema

from ckanext.files import types  # isort: skip # noqa: F401


_actions, action = make_collector()


def get_actions():
    return dict(_actions)


def _flat_mask(data):
    # type: (dict[str, types.Any]) -> dict[tuple[types.Any], types.Any]
    result = {}  # type: dict[tuple[types.Any], types.Any]

    for k, v in data.items():
        if isinstance(v, dict):
            result.update({(k,) + sk: sv for sk, sv in _flat_mask(v).items()})
        else:
            result[(k,)] = v

    return result


@action
@tk.side_effect_free
@validate(schema.file_search_by_user)
def files_file_search_by_user(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> dict[str, types.Any]
    tk.check_access("files_file_search_by_user", context, data_dict)
    sess = context["session"]

    user = model.User.get(data_dict.get("user", context["user"]))
    if not user:
        raise tk.ObjectNotFound("user")

    q = sess.query(File).join(
        Owner,
        sa.and_(File.id == Owner.item_id, Owner.item_type == "file"),
    )

    if "storage" in data_dict:
        q = q.filter(File.storage == data_dict["storage"])

    q = q.filter(sa.and_(Owner.owner_type == "user", Owner.owner_id == user.id))

    inspector = sa.inspect(File)  # type: types.Any
    columns = inspector.columns

    for mask in ["storage_data", "plugin_data"]:
        if mask in data_dict:
            for k, v in _flat_mask(data_dict[mask]).items():
                field = columns[mask]
                for segment in k:
                    field = field[segment]

                q = q.filter(field.astext == v)

    total = q.count()

    parts = data_dict["sort"].split(".")
    sort = parts[0]
    sort_path = parts[1:]

    if sort not in columns:
        raise tk.ValidationError({"sort": ["Unknown sort column"]})

    column = columns[sort]

    if sort_path and sort == "storage_data":
        for part in sort_path:
            column = column[part]

    if data_dict["reverse"]:
        column = column.desc()

    q = q.order_by(column)

    q = q.limit(data_dict["rows"]).offset(data_dict["start"])

    return {"count": total, "results": [f.dictize(context) for f in q]}


@action
@validate(schema.file_create)
def files_file_create(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> dict[str, types.Any]
    tk.check_access("files_file_create", context, data_dict)
    _ensure_name(data_dict)

    extras = data_dict.get("__extras", {})

    try:
        storage = shared.get_storage(data_dict["storage"])
    except exceptions.UnknownStorageError as err:
        raise tk.ValidationError({"storage": [str(err)]})  # noqa: B904

    if not storage.supports(Capability.CREATE):
        raise tk.ValidationError({"storage": ["Operation is not supported"]})

    try:
        storage_data = storage.upload(
            secure_filename(data_dict["name"]),
            data_dict["upload"],
            extras,
        )
    except exceptions.UploadError as err:
        raise tk.ValidationError({"upload": [str(err)]})  # noqa: B904

    fileobj = File(
        name=data_dict["name"],
        storage=data_dict["storage"],
        storage_data=storage_data,
        completed=True,
    )
    context["session"].add(fileobj)
    _set_owner(context, "file", fileobj.id)  # type: ignore
    context["session"].commit()

    return fileobj.dictize(context)


def _ensure_name(data_dict, name_field="name", upload_field="upload"):
    # type: (dict[str, types.Any], str, str) -> None
    if name_field in data_dict:
        return
    name = data_dict[upload_field].filename
    if not name:
        raise tk.ValidationError(
            {
                name_field: [
                    "Name is missing and cannot be deduced from {}".format(
                        upload_field,
                    ),
                ],
            },
        )
    data_dict[name_field] = name


def _set_owner(context, item_type, item_id):
    # type: (types.Any, str, str) -> None
    user = model.User.get(context["user"])
    if user:
        owner = Owner(
            item_id=item_id,
            item_type=item_type,
            owner_id=user.id,
            owner_type="user",
        )
        context["session"].add(owner)


def _delete_owner(context, item_type, item_id):
    # type: (types.Any, str, str) -> None
    stmt = sa.delete(Owner).where(
        sa.and_(
            Owner.item_type == item_type,
            Owner.item_id == item_id,
        ),
    )
    context["session"].execute(stmt)


@action
@validate(schema.file_delete)
def files_file_delete(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> bool
    tk.check_access("files_file_delete", context, data_dict)

    data_dict["id"]
    fileobj = context["session"].query(File).filter_by(id=data_dict["id"]).one_or_none()
    if not fileobj:
        raise tk.ObjectNotFound("file")

    storage = shared.get_storage(fileobj.storage)
    if not storage.supports(Capability.REMOVE):
        raise tk.ValidationError({"storage": ["Operation is not supported"]})

    try:
        storage.remove(fileobj.storage_data)
    except exceptions.PermissionError as err:
        raise tk.NotAuthorized(str(err))  # noqa: B904

    _delete_owner(context, "file", fileobj.id)
    context["session"].delete(fileobj)
    context["session"].commit()

    return fileobj.dictize(context)


@action
@tk.side_effect_free
@validate(schema.file_show)
def files_file_show(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> dict[str, types.Any]
    tk.check_access("files_file_show", context, data_dict)

    fileobj = (
        context["session"].query(File).filter(File.id == data_dict["id"]).one_or_none()
    )
    if not fileobj:
        raise tk.ObjectNotFound("file")

    return fileobj.dictize(context)


@action
@validate(schema.file_rename)
def files_file_rename(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> dict[str, types.Any]
    tk.check_access("files_file_rename", context, data_dict)

    fileobj = (
        context["session"].query(File).filter(File.id == data_dict["id"]).one_or_none()
    )  # type: File | None
    if not fileobj:
        raise tk.ObjectNotFound("file")

    fileobj.name = data_dict["name"]
    fileobj.touch()
    context["session"].commit()

    return fileobj.dictize(context)


@action
@validate(schema.upload_initialize)
def files_upload_initialize(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> dict[str, types.Any]

    tk.check_access("files_upload_initialize", context, data_dict)
    _ensure_name(data_dict)
    extras = data_dict.get("__extras", {})

    try:
        storage = shared.get_storage(data_dict["storage"])
    except exceptions.UnknownStorageError as err:
        raise tk.ValidationError({"storage": [str(err)]})  # noqa: B904

    if not storage.supports(Capability.MULTIPART_UPLOAD):
        raise tk.ValidationError({"storage": ["Operation is not supported"]})

    try:
        storage_data = storage.initialize_multipart_upload(
            secure_filename(data_dict["name"]),
            extras,
        )
    except exceptions.UploadError as err:
        raise tk.ValidationError({"upload": [str(err)]})  # noqa: B904

    fileobj = File(
        name=data_dict["name"],
        storage=data_dict["storage"],
        storage_data=storage_data,
    )
    context["session"].add(fileobj)
    _set_owner(context, "file", fileobj.id)  # type: ignore
    context["session"].commit()

    return fileobj.dictize(context)


@action
@tk.side_effect_free
@validate(schema.upload_show)
def files_upload_show(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> dict[str, types.Any]
    tk.check_access("files_upload_show", context, data_dict)
    file_dict = tk.get_action("files_file_show")(context, data_dict)

    if file_dict["completed"]:
        raise tk.ObjectNotFound("upload")

    storage = shared.get_storage(file_dict["storage"])
    storage_data = storage.show_multipart_upload(file_dict["storage_data"])

    return dict(file_dict, storage_data=storage_data)


@action
@validate(schema.upload_update)
def files_upload_update(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> dict[str, types.Any]
    tk.check_access("files_upload_update", context, data_dict)

    extras = data_dict.get("__extras", {})

    fileobj = context["session"].query(File).filter_by(id=data_dict["id"]).one_or_none()
    if not fileobj:
        raise tk.ObjectNotFound("upload")

    storage = shared.get_storage(fileobj.storage)

    try:
        fileobj.storage_data = storage.update_multipart_upload(fileobj.storage_data, extras)
    except exceptions.UploadError as err:
        raise tk.ValidationError({"upload": [str(err)]})

    context["session"].commit()

    return fileobj.dictize(context)


@action
@validate(schema.upload_complete)
def files_upload_complete(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> dict[str, types.Any]
    tk.check_access("files_upload_complete", context, data_dict)

    extras = data_dict.get("__extras", {})

    data_dict["id"]
    fileobj = context["session"].query(File).filter_by(id=data_dict["id"]).one_or_none()
    if not fileobj:
        raise tk.ObjectNotFound("upload")

    storage = shared.get_storage(fileobj.storage)

    try:
        fileobj.storage_data = storage.complete_multipart_upload(
            fileobj.storage_data,
            extras,
        )
    except exceptions.UploadError as err:
        raise tk.ValidationError({"upload": [str(err)]})

    fileobj.completed = True
    context["session"].commit()

    return fileobj.dictize(context)



from datetime import datetime
from io import BytesIO

import pytest
import six
from freezegun import freeze_time
from responses import RequestsMock
from werkzeug.datastructures import FileStorage

import ckan.plugins.toolkit as tk
from ckan.lib.redis import connect_to_redis
from ckan.tests.helpers import call_action

if six.PY3:
    from typing import Any  # isort: skip # noqa: F401


@pytest.fixture()
def responses(ckan_config):
    # type: (Any) -> Any
    with RequestsMock() as rsps:
        rsps.add_passthru(ckan_config["solr_url"])
        yield rsps


@pytest.fixture()
def files_stopped_time():
    now = datetime.now()
    with freeze_time(now):
        yield now


if tk.check_ckan_version("2.9"):

    @pytest.fixture
    def clean_db(reset_db, migrate_db_for):
        # type: (Any, Any) -> None

        # fix for CKAN v2.9 issue with `reset_db` attempting to remove all
        # registered models, not only core-models
        migrate_db_for("files")
        reset_db()
        migrate_db_for("files")

else:

    @pytest.fixture
    def clean_db(reset_db):
        # type: (Any) -> None
        from ckanext.files.command import create_tables

        reset_db()
        create_tables()


class FakeFileStorage(FileStorage):
    def __init__(self, stream, filename):
        # type: (Any, str) -> None
        super(FakeFileStorage, self).__init__(stream, filename, "upload")


@pytest.fixture
def create_with_upload(ckan_config, monkeypatch, tmpdir):
    # type: (Any, Any, Any) -> Any
    """Reimplementation of original CKAN fixture with better fake storage.

    CKAN version adds just a few attributes to FileStorage object, while
    current plugin requires exact immitation.
    """

    monkeypatch.setitem(ckan_config, "ckan.storage_path", str(tmpdir))

    def factory(data, filename, context=None, **kwargs):
        # type: (Any, Any, Any, **Any) -> Any
        if context is None:
            context = {}

        action = kwargs.pop("action", "resource_create")
        field = kwargs.pop("upload_field_name", "upload")
        test_file = BytesIO()
        if isinstance(data, six.text_type):
            data = data.encode()
        test_file.write(data)
        test_file.seek(0)
        test_resource = FakeFileStorage(test_file, filename)

        params = {
            field: test_resource,
        }
        params.update(kwargs)
        return call_action(action, context, **params)

    return factory


@pytest.fixture(scope="session")
def reset_redis():
    def cleaner(pattern="*"):
        # type: (str) -> int
        """Remove keys matching pattern.
        Return number of removed records.
        """
        conn = connect_to_redis()
        keys = conn.keys(pattern)
        if keys:
            return conn.delete(*keys)
        return 0

    return cleaner


@pytest.fixture()
def clean_redis(reset_redis):
    """Remove all keys from Redis.
    This fixture removes all the records from Redis::
        @pytest.mark.usefixtures("clean_redis")
        def test_redis_is_empty():
            assert redis.keys("*") == []
    If test requires presence of some initial data in redis, make sure that
    data producer applied **after** ``clean_redis``::
        @pytest.mark.usefixtures(
            "clean_redis",
            "fixture_that_adds_xxx_key_to_redis"
        )
        def test_redis_has_one_record():
            assert redis.keys("*") == [b"xxx"]
    """
    reset_redis()



"""Exception definitions for the extension.

Avoid raising python-native exceptions and prefere defining `FilesError`
subclass.

"""

import six

from ckan.lib.formatters import localised_filesize

if six.PY3:
    from typing import Any  # isort: skip # noqa: F401


class FilesError(Exception):
    pass


class StorageError(FilesError):
    pass


class UploadError(StorageError):
    pass


class UnknownStorageError(StorageError):
    """Storage with the given name is not configured."""

    def __init__(self, storage):
        # type: (str) -> None
        self.storage = storage

    def __str__(self):
        return "Storage {} is not configured".format(self.storage)


class UnknownAdapterError(StorageError):
    """Specified storage adapter is not registered."""

    def __init__(self, adapter):
        # type: (str) -> None
        self.adapter = adapter

    def __str__(self):
        return "Storage adapter {} is not registered".format(self.adapter)


class UnsupportedOperationError(StorageError):
    """Requested operation is not supported by storage."""

    def __init__(self, operation, adapter):
        # type: (Any, Any) -> None
        self.operation = operation
        self.adapter = adapter

    def __str__(self):
        return "Operation {} is not supported by storage adapter {}".format(
            self.operation,
            self.adapter,
        )


class InvalidStorageConfigurationError(StorageError):
    """Storage cannot be initialized with given configuration."""

    def __init__(self, adapter, problem):
        # type: (type, str) -> None
        self.adapter = adapter
        self.problem = problem

    def __str__(self):
        return "Cannot initialize storage adapter {} due to following error: {}".format(
            self.adapter.__name__,
            self.problem,
        )


class PermissionError(StorageError):
    """Storage client does not have required permissions."""

    def __init__(self, adapter, operation, problem):
        # type: (type, str, str) -> None
        self.adapter = adapter
        self.operation = operation
        self.problem = problem

    def __str__(self):
        msg = "Storage {} is not allowed to perform {} operation: {}"
        return msg.format(
            self.adapter.__name__,
            self.operation,
            self.problem,
        )


class MissingStorageConfigurationError(InvalidStorageConfigurationError):
    """Storage cannot be initialized due to missing option."""

    def __init__(self, adapter, option):
        # type: (type, str) -> None
        return super(MissingStorageConfigurationError, self).__init__(
            adapter,
            "{} option is required".format(option),
        )


class MissingFileError(StorageError):
    """File does not exist."""

    def __init__(self, storage, filename):
        # type: (str, str) -> None
        self.storage = storage
        self.filename = filename

    def __str__(self):
        return "File {} does not exist inside storage {}".format(
            self.filename,
            self.storage,
        )


class ExistingFileError(StorageError):
    """File already exists."""

    def __init__(self, storage, filename):
        # type: (str, str) -> None
        self.storage = storage
        self.filename = filename

    def __str__(self):
        return "File {} already exists inside storage {}".format(
            self.filename,
            self.storage,
        )


class LargeUploadError(UploadError):
    """Storage cannot be initialized due to missing option."""

    def __init__(self, actual_size, max_size):
        # type: (int, int) -> None
        self.actual_size = actual_size
        self.max_size = max_size

    def __str__(self):
        return "Upload size {} surpasses max allowed size {}".format(
            localised_filesize(self.actual_size),
            localised_filesize(self.max_size),
        )


class UploadOutOfBoundError(LargeUploadError):
    """Multipart upload exceeds expected size."""

    def __str__(self):
        return "Upload size {} exceeds expected size {}".format(
            localised_filesize(self.actual_size),
            localised_filesize(self.max_size),
        )


class NameStrategyError(UploadError):
    """Undefined name strategy."""

    def __init__(self, strategy):
        # type: (str) -> None
        self.strategy = strategy

    def __str__(self):
        return "Unknown name strategy {}".format(self.strategy)


class UploadExtrasError(UploadError):
    """Wrong extras passed during upload."""

    def __init__(self, extras):
        # type: (Any) -> None
        self.extras = extras

    def __str__(self):
        return "Wrong extras: {}".format(self.extras)


class MissingExtrasError(UploadExtrasError):
    """Wrong extras passed during upload."""

    def __init__(self, key):
        # type: (Any) -> None
        self.key = key

    def __str__(self):
        return "Key {} is missing from upload extras".format(self.key)



from __future__ import annotations

from typing import Any

from dominate import tags

import ckan.plugins.toolkit as tk

from ckanext.ap_main.collection.base import (
    ApCollection,
    ApColumns,
    ApHtmxTableSerializer,
    BulkAction,
    RowAction,
)
from ckanext.collection.types import ButtonFilter, InputFilter
from ckanext.collection.utils import Filters
from ckanext.collection.utils.data.model import ModelData

from ckanext.files.model import File


def file_row_dictizer(serializer: ApHtmxTableSerializer[Any], row: File):
    data = row.dictize({})
    data["bulk-action"] = data["id"]

    return data


class FileManagerCollection(ApCollection):
    SerializerFactory = ApHtmxTableSerializer.with_attributes(
        record_template="file_manager/record.html",
        row_dictizer=file_row_dictizer,
    )

    ColumnsFactory = ApColumns.with_attributes(
        names=[
            "bulk-action",
            "name",
            "path",
            "kind",
            "uploaded_at",
            "last_access",
            "extras",
            "row_actions",
        ],
        sortable={"name", "kind", "uploaded_at", "last_access"},
        searchable={"name"},
        labels={
            "bulk-action": tk.literal(
                tags.input_(
                    type="checkbox",
                    name="bulk_check",
                    id="bulk_check",
                    data_module="ap-bulk-check",
                    data_module_selector='input[name="entity_id"]',
                ),
            ),
            "name": "Name",
            "path": "Path",
            "kind": "Type",
            "uploaded_at": "Uploaded At",
            "last_access": "Last Access",
            "extras": "Extras",
            "row_actions": "Actions",
        },
        width={"name": "20%", "path": "20%"},
        serializers={
            "uploaded_at": [("date", {})],
            "extras": [("json_display", {})],
            "last_access": [("day_passed", {})],
        },
    )

    DataFactory = ModelData.with_attributes(
        model=File,
        is_scalar=True,
        use_naive_search=True,
        use_naive_filters=True,
    )

    FiltersFactory = Filters.with_attributes(
        static_actions=[
            BulkAction(
                name="bulk-action",
                type="bulk_action",
                options={
                    "label": "Action",
                    "options": [
                        {"value": "1", "text": "Remove selected files"},
                    ],
                },
            ),
            RowAction(
                name="edit",
                type="row_action",
                options={
                    "endpoint": "",
                    "label": "",
                    "icon": "fa fa-pencil",
                    "params": {
                        "data-module-path": "$id",
                        "view": "edit",
                    },
                },
            ),
            RowAction(
                name="view",
                type="row_action",
                options={
                    "endpoint": "ap_content.entity_proxy",
                    "label": "View",
                    "params": {
                        "entity_id": "$id",
                        "entity_type": "$type",
                        "view": "read",
                    },
                },
            ),
        ],
        static_filters=[
            InputFilter(
                name="q",
                type="input",
                options={
                    "label": "Search",
                    "placeholder": "Search",
                },
            ),
            ButtonFilter(
                name="type",
                type="button",
                options={
                    "label": "Clear",
                    "type": "button",
                    "attrs": {
                        "onclick": (
                            "$(this).closest('form').find('input,select')"
                            ".val('').prevObject[0].requestSubmit()"
                        ),
                    },
                },
            ),
        ],
    )



import logging
import os

import magic
from werkzeug.datastructures import FileStorage

import ckan.plugins.toolkit as tk

from ckanext.files import exceptions, types, utils
from ckanext.files.base import (
    Capability,
    HashingReader,
    Manager,
    Reader,
    Storage,
    Uploader,
)

FsAdditionalData = types.TypedDict("FsAdditionalData", {})


class FsStorageData(FsAdditionalData, types.MinimalStorageData):
    pass


log = logging.getLogger(__name__)
CHUNK_SIZE = 16384


class FileSystemUploader(Uploader):
    required_options = ["path"]
    capabilities = utils.combine_capabilities(
        Capability.CREATE,
        Capability.MULTIPART_UPLOAD,
    )

    def upload(self, name, upload, extras):
        # type: (str, types.Upload, dict[str, types.Any]) -> FsStorageData
        filename = self.storage.compute_name(name, extras, upload)
        filepath = os.path.join(self.storage.settings["path"], filename)

        os.makedirs(os.path.dirname(filepath), exist_ok=True)

        reader = HashingReader(upload.stream)
        with open(filepath, "wb") as dest:
            for chunk in reader:
                dest.write(chunk)

        return {
            "filename": filename,
            "content_type": upload.content_type,
            "size": os.path.getsize(filepath),
            "hash": reader.get_hash(),
        }

    def initialize_multipart_upload(self, name, extras):
        # type: (str, dict[str, types.Any]) -> dict[str, types.Any]
        schema = {
            "size": [
                tk.get_validator("not_missing"),
                tk.get_validator("int_validator"),
            ],
            "__extras": [tk.get_validator("ignore")],
        }  # type: dict[str, types.Any]
        data, errors = tk.navl_validate(extras, schema)

        if errors:
            raise tk.ValidationError(errors)

        upload = FileStorage(content_length=data["size"])

        max_size = self.storage.max_size
        if max_size:
            utils.ensure_size(upload, max_size)

        result = dict(self.upload(name, upload, data))
        result["size"] = data["size"]
        result["content_type"] = "application/octet-stream"
        result["uploaded"] = 0

        return result

    def show_multipart_upload(self, upload_data):
        # type: (dict[str, types.Any]) -> dict[str, types.Any]
        return upload_data

    def update_multipart_upload(self, upload_data, extras):
        # type: (dict[str, types.Any], dict[str, types.Any]) -> dict[str, types.Any]
        schema = {
            "position": [
                tk.get_validator("ignore_missing"),
                tk.get_validator("int_validator"),
            ],
            "upload": [
                tk.get_validator("not_missing"),
                tk.get_validator("files_into_upload"),
            ],
            "__extras": [tk.get_validator("ignore")],
        }
        data, errors = tk.navl_validate(extras, schema)

        if errors:
            raise tk.ValidationError(errors)

        data.setdefault("position", upload_data["uploaded"])
        upload = data["upload"]  # type: types.Upload

        expected_size = data["position"] + upload.content_length
        if expected_size > upload_data["size"]:
            raise exceptions.UploadOutOfBoundError(expected_size, upload_data["size"])

        filepath = os.path.join(
            str(self.storage.settings["path"]),
            upload_data["filename"],
        )
        with open(filepath, "rb+") as dest:
            dest.seek(data["position"])
            dest.write(upload.stream.read())

        upload_data["uploaded"] = os.path.getsize(filepath)
        return upload_data

    def complete_multipart_upload(self, upload_data, extras):
        # type: (dict[str, types.Any], dict[str, types.Any]) -> FsStorageData
        filepath = os.path.join(
            str(self.storage.settings["path"]),
            upload_data["filename"],
        )
        size = os.path.getsize(filepath)
        if size != upload_data["size"]:
            raise tk.ValidationError(
                {
                    "size": [
                        "Actual filesize {} does not match expected {}".format(
                            size,
                            upload_data["size"],
                        ),
                    ],
                },
            )

        with open(filepath, "rb") as src:
            reader = HashingReader(src)
            content_type = magic.from_buffer(next(reader, b""), True)
            reader.exhaust()

        return {
            "filename": upload_data["filename"],
            "content_type": content_type,
            "size": size,
            "hash": reader.get_hash(),
        }


class FileSystemManager(Manager):
    required_options = ["path"]
    capabilities = utils.combine_capabilities(Capability.REMOVE, Capability.ANALYZE)

    def remove(self, data):
        # type: (dict[str, types.Any]) -> bool
        filepath = os.path.join(str(self.storage.settings["path"]), data["filename"])
        if not os.path.exists(filepath):
            return False

        os.remove(filepath)
        return True

    def analyze(self, filename):
        # type: (str) -> types.FsStorageData
        """Return all details about filename."""
        filepath = os.path.join(str(self.storage.settings["path"]), filename)
        if not os.path.exists(filepath):
            raise exceptions.MissingFileError(self.storage.settings["name"], filepath)

        with open(filepath, "rb") as src:
            reader = HashingReader(src)
            content_type = magic.from_buffer(next(reader, b""), True)
            reader.exhaust()

        return {
            "filename": filename,
            "content_type": content_type,
            "size": os.path.getsize(filepath),
            "hash": reader.get_hash(),
        }


class FileSystemReader(Reader):
    required_options = ["path"]
    capabilities = utils.combine_capabilities(Capability.STREAM)

    def stream(self, data):
        # type: (dict[str, types.Any]) -> types.IO[bytes]
        filepath = os.path.join(str(self.storage.settings["path"]), data["filename"])
        if not os.path.exists(filepath):
            raise exceptions.MissingFileError(self.storage.settings["name"], filepath)

        return open(filepath, "rb")


class FileSystemStorage(Storage):
    def make_uploader(self):
        return FileSystemUploader(self)

    def make_reader(self):
        return FileSystemReader(self)

    def make_manager(self):
        return FileSystemManager(self)

    def __init__(self, **settings):
        # type: (**types.Any) -> None
        path = self.ensure_option(settings, "path")

        if not os.path.exists(path):
            if tk.asbool(settings.get("create_path")):
                os.makedirs(path)
            else:
                raise exceptions.InvalidStorageConfigurationError(
                    type(self),
                    "path `{}` does not exist".format(path),
                )

        super(FileSystemStorage, self).__init__(**settings)

    @classmethod
    def declare_config_options(cls, declaration, key):  # pragma: no cover
        # type: (types.Declaration, types.Key) -> None
        super().declare_config_options(declaration, key)
        declaration.declare(key.path).required().set_description(
            "Path to the folder where uploaded data will be stored.",
        )
        declaration.declare_bool(key.create_path).set_description(
            "Create storage folder if it does not exist.",
        )


class PublicFileSystemStorage(FileSystemStorage):
    def __init__(self, **settings):
        # type: (**types.Any) -> None
        self.ensure_option(settings, "public_root")
        super(PublicFileSystemStorage, self).__init__(**settings)

    @classmethod
    def declare_config_options(cls, declaration, key):  # pragma: no cover
        # type: (types.Declaration, types.Key) -> None
        super().declare_config_options(declaration, key)
        declaration.declare(key.public_root).required().set_description(
            "URL of the storage folder."
            + " `public_root + filename` must produce a public URL",
        )



import os

import click

import ckan.plugins.toolkit as tk
from ckan import model

from ckanext.files import base, config, exceptions, shared
from ckanext.files.model import File, Owner

__all__ = [
    "files",
]


def get_commands():
    return [files]


@click.group(short_help="ckanext-files CLI commands")
def files():
    pass


@files.command()
def adapters():
    """Show all awailable storage adapters."""
    for name in sorted(base.adapters):
        adapter = base.adapters.get(name)
        click.secho(
            "{} - {}:{}".format(
                click.style(name, bold=True),
                adapter.__module__,
                adapter.__name__,
            ),
        )


@files.command()
def storages():
    """Show all configured storages."""
    for name, settings in config.storages().items():
        click.secho("{}: {}".format(click.style(name, bold=True), settings["type"]))


@files.command()
@click.option("-s", "--storage-name", help="Name of the configured storage")
@click.option(
    "-o",
    "--orphans-only",
    help="Show only files not registered in DB",
    is_flag=True,
)
@click.option("-m", "--materialize", help="Register orphans in DB", is_flag=True)
@click.option("-a", "--adopt-by", help="Attach orphans to specified user")
def scan(storage_name, orphans_only, adopt_by, materialize):
    # type: (str | None, bool, str | None, bool) -> None
    """Iterate over all files available in storage.

    This command can be used to locate "orphans", that are not registered in
    DB, but exist in storage.

    """
    storage_name = storage_name or config.default_storage()
    storage = shared.get_storage(storage_name)

    try:
        files = storage.scan()
    except exceptions.UnsupportedOperationError as err:
        tk.error_shout(err)
        raise click.Abort()  # noqa: B904

    stepfather = model.User.get(adopt_by)

    for name in files:
        is_orphan = not model.Session.query(
            File.by_location(name, storage_name).exists(),
        ).scalar()

        if orphans_only and not is_orphan:
            continue

        click.echo(name)

        if materialize and is_orphan:
            try:
                data = storage.analyze(name)
            except exceptions.UnsupportedOperationError as err:
                tk.error_shout(err)
                raise click.Abort()  # noqa: B904

            fileobj = File(
                name=os.path.basename(name),
                storage=storage_name,
                storage_data=data,
                completed=True,
            )
            model.Session.add(fileobj)

            if stepfather:
                owner = Owner(
                    item_id=fileobj.id,
                    item_type="file",
                    owner_id=stepfather.id,
                    owner_type="user",
                )
                model.Session.add(owner)

            model.Session.commit()



"""add access column to owner table

Revision ID: 76fdef67f479
Revises: 64ca007bf3eb
Create Date: 2024-03-18 16:17:35.734318

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "76fdef67f479"
down_revision = "64ca007bf3eb"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("files_owner", sa.Column("access", sa.Text, server_default="'full'"))


def downgrade():
    op.drop_column("files_owner", "access")



import hashlib
import uuid
from io import BytesIO

import pytest
from werkzeug.datastructures import FileStorage

from ckanext.files import base, exceptions, utils
from ckanext.files.storage import RedisStorage

from datetime import datetime  # isort: skip # noqa: F401


from faker import Faker  # isort: skip # noqa: F401


class TestStorageFromSettings:
    def test_missing_adapter(self):
        """Unknown udapter causes an exception."""

        with pytest.raises(exceptions.UnknownAdapterError):
            base.storage_from_settings("", {})

    def test_invalid_configuration(self):
        """Wrong configuration causes an exception."""

        with pytest.raises(exceptions.InvalidStorageConfigurationError):
            base.storage_from_settings("", {"type": "files:fs"})

    def test_normal_configuration(self):
        """Valid configuration produces a storage."""

        storage = base.storage_from_settings("", {"type": "files:redis"})
        assert isinstance(storage, RedisStorage)


class TestHasingReader:
    def test_empty_hash(self):
        """Empty reader produces the hash of empty string and doesn't add any
        default bytes.

        """

        reader = base.HashingReader(BytesIO())
        reader.exhaust()

        assert reader.get_hash() == hashlib.md5().hexdigest()

    def test_hash(self, faker):
        # type: (Faker) -> None
        """Reader's hash is based on the stream content."""

        content = faker.binary(100)
        expected = hashlib.md5(content).hexdigest()

        reader = base.HashingReader(BytesIO(content))

        output = b""
        for chunk in reader:
            output += chunk

        assert output == content
        assert reader.get_hash() == expected

    def test_reset(self, faker):
        # type: (Faker) -> None
        """Resetting the reader makes it reusable"""
        stream = BytesIO(faker.binary(100))
        reader = base.HashingReader(stream)

        reader.exhaust()

        first_hash = reader.get_hash()
        assert stream.tell() == 100

        reader.reset()
        assert stream.tell() == 0

        reader.exhaust()
        assert reader.get_hash() == first_hash


class TestOptionCheckeer:
    def test_missing_option(self, faker):
        # type: (Faker) -> None
        """Checker raises an exception if option is missing."""

        with pytest.raises(exceptions.MissingStorageConfigurationError):
            base.OptionChecker.ensure_option({faker.word(): True}, faker.word())

    def test_existing_option(self, faker):
        # type: (Faker) -> None
        """Checker returns the value of existing option"""
        option = faker.word()
        value = faker.word()

        result = base.OptionChecker.ensure_option({option: value}, option)
        assert result == value


class TestStorageService:
    @pytest.fixture()
    def service_class(self):
        class Service(base.StorageService):
            required_options = ["test"]

        return Service

    def test_missing_option(self, service_class):
        # type: (type[base.StorageService]) -> None
        """Service cannot be initialized without required option."""
        storage = base.Storage()
        with pytest.raises(exceptions.MissingStorageConfigurationError):
            service_class(storage)

    def test_existing_option(self, service_class):
        # type: (type[base.StorageService]) -> None
        """Service works with required options."""
        storage = base.Storage(test=1)
        service_class(storage)


class TestUploader:
    @pytest.fixture()
    def uploader(self):
        return base.Uploader(base.Storage())

    def test_abstract_methods(self, uploader, faker):
        # type: (base.Uploader, Faker) -> None
        """Abstract methods raise exception."""

        with pytest.raises(NotImplementedError):
            uploader.upload(faker.file_name(), FileStorage(), {})

        with pytest.raises(NotImplementedError):
            uploader.initialize_multipart_upload(faker.file_name(), {})

        with pytest.raises(NotImplementedError):
            uploader.show_multipart_upload({})

        with pytest.raises(NotImplementedError):
            uploader.update_multipart_upload({}, {})

        with pytest.raises(NotImplementedError):
            uploader.complete_multipart_upload({}, {})


class TestManager:
    @pytest.fixture()
    def manager(self):
        return base.Manager(base.Storage())

    def test_abstract_methods(self, manager):
        # type: (base.Manager) -> None
        """Abstract methods raise exception."""

        with pytest.raises(NotImplementedError):
            manager.remove({})


class TestReader:
    @pytest.fixture()
    def reader(self):
        return base.Reader(base.Storage())

    def test_abstract_methods(self, reader):
        # type: (base.Reader) -> None
        """Abstract methods raise exception."""

        with pytest.raises(NotImplementedError):
            reader.stream({})


class RemovingManager(base.Manager):
    capabilities = utils.combine_capabilities(base.Capability.REMOVE)


class StreamingReader(base.Reader):
    capabilities = utils.combine_capabilities(base.Capability.STREAM)


class SimpleUploader(base.Uploader):
    capabilities = utils.combine_capabilities(base.Capability.CREATE)


class Storage(base.Storage):
    def make_reader(self):
        return StreamingReader(self)

    def make_uploader(self):
        return SimpleUploader(self)

    def make_manager(self):
        return RemovingManager(self)


class TestStorage:
    def test_inherited_capabilities(self):
        """Storage combine capabilities of its services."""
        storage = Storage()
        assert storage.capabilities == utils.combine_capabilities(
            base.Capability.REMOVE,
            base.Capability.STREAM,
            base.Capability.CREATE,
        )

    def test_settings(self, faker):
        # type: (Faker) -> None
        """Storage keeps all incoming arguments as settings."""

        settings = faker.pydict()
        storage = base.Storage(**settings)
        assert storage.settings == settings

    def test_max_size(self, faker):
        # type: (Faker) -> None
        """Storage has a dedicated property for `max_size` setting."""

        assert base.Storage().max_size == 0

        max_size = faker.pyint()
        assert base.Storage(max_size=max_size).max_size == max_size

    def test_supports(self):
        """Storage can tell whether it supports certain capabilities."""

        storage = Storage()

        assert storage.supports(base.Capability.CREATE)
        assert storage.supports(
            utils.combine_capabilities(
                base.Capability.REMOVE,
                base.Capability.STREAM,
            ),
        )

        assert not storage.supports(base.Capability.MULTIPART_UPLOAD)
        assert not storage.supports(
            utils.combine_capabilities(
                base.Capability.REMOVE,
                base.Capability.MULTIPART_UPLOAD,
            ),
        )

    def test_not_supported_methods(self, faker):
        # type: (Faker) -> None
        with pytest.raises(exceptions.UnsupportedOperationError):
            base.Storage().upload(faker.file_name(), FileStorage(), {})

        with pytest.raises(exceptions.UnsupportedOperationError):
            base.Storage().stream({})

        with pytest.raises(exceptions.UnsupportedOperationError):
            base.Storage().remove({})

        with pytest.raises(exceptions.UnsupportedOperationError):
            base.Storage().copy({}, base.Storage(), "", {})

        with pytest.raises(exceptions.UnsupportedOperationError):
            base.Storage().move({}, base.Storage(), "", {})

    def test_upload_checks_max_size(self, faker):
        # type: (Faker) -> None
        """Storage raises an error if upload exceeds max size."""
        storage = Storage(max_size=10)
        with pytest.raises(exceptions.LargeUploadError):
            storage.upload(
                faker.file_name(),
                FileStorage(BytesIO(faker.binary(20))),
                {},
            )

    def test_not_implemented_methods(self, faker):
        # type: (Faker) -> None
        """Storage raises an error if upload is not implemented."""
        storage = Storage()
        with pytest.raises(NotImplementedError):
            storage.upload(faker.file_name(), FileStorage(), {})

        with pytest.raises(NotImplementedError):
            storage.initialize_multipart_upload(faker.file_name(), {})

        with pytest.raises(NotImplementedError):
            storage.show_multipart_upload({})

        with pytest.raises(NotImplementedError):
            storage.update_multipart_upload({}, {})

        with pytest.raises(NotImplementedError):
            storage.complete_multipart_upload({}, {})

        with pytest.raises(NotImplementedError):
            storage.remove({})

        with pytest.raises(NotImplementedError):
            storage.copy({}, storage, "", {})

        with pytest.raises(NotImplementedError):
            storage.move({}, storage, "", {})

    def test_compute_name_uuid(self, faker):
        # type: (Faker) -> None
        """`uuid`(default) name strategy produces valid UUID."""
        storage = Storage()

        extension = faker.file_extension()
        name = faker.file_name(extension=extension)
        result = storage.compute_name(name, {})

        assert uuid.UUID(result)

    def test_compute_name_uuid_prefix(self, faker):
        # type: (Faker) -> None
        """`uuid_prefix` name strategy produces valid UUID."""
        storage = Storage()

        storage.settings["name_strategy"] = "uuid_prefix"
        extension = faker.file_extension()
        name = faker.file_name(extension=extension)
        result = storage.compute_name(name, {})
        assert result.endswith(name)
        assert uuid.UUID(result[: -len(name)])

    def test_compute_name_uuid_with_extension(self, faker):
        # type: (Faker) -> None
        """`uuid_with_extension` name strategy produces valid UUID."""
        storage = Storage()
        storage.settings["name_strategy"] = "uuid_with_extension"
        extension = faker.file_extension()
        name = faker.file_name(extension=extension)
        result = storage.compute_name(name, {})

        assert result.endswith(extension)
        assert uuid.UUID(result[: -len(extension) - 1])

    def test_compute_name_datetime_prefix(self, faker, files_stopped_time):
        # type: (Faker, datetime) -> None
        """`datetime_prefix` name strategy produces valid UUID."""
        storage = Storage()
        storage.settings["name_strategy"] = "datetime_prefix"
        extension = faker.file_extension()
        name = faker.file_name(extension=extension)
        result = storage.compute_name(name, {})

        assert result == files_stopped_time.isoformat() + name

    def test_compute_name_datetime_with_extension(
        self,
        faker,
        files_stopped_time,
    ):
        # type: (Faker, datetime) -> None
        """`datetime_with_extension` name strategy produces valid UUID."""
        storage = Storage()
        storage.settings["name_strategy"] = "datetime_with_extension"
        extension = faker.file_extension()
        name = faker.file_name(extension=extension)
        result = storage.compute_name(name, {})

        assert result == files_stopped_time.isoformat() + "." + extension

    def test_compute_name_with_wrong_strategy(self):
        # type: () -> None
        """`datetime_with_extension` name strategy produces valid UUID."""
        storage = Storage()
        storage.settings["name_strategy"] = "wrong_strategy"
        with pytest.raises(exceptions.NameStrategyError):
            storage.compute_name("test", {})



from datetime import datetime

from sqlalchemy.ext.declarative import declarative_base

from ckan.model.meta import metadata

Base = declarative_base(metadata=metadata)


def now():
    return datetime.utcnow()



from io import BytesIO

import ckan.plugins.toolkit as tk
from ckan.lib.redis import connect_to_redis  # type: ignore

from ckanext.files import exceptions, types, utils
from ckanext.files.base import (
    Capability,
    HashingReader,
    Manager,
    Reader,
    Storage,
    Uploader,
)

import redis  # isort: skip # noqa: F401


RedisAdditionalData = types.TypedDict("RedisAdditionalData", {})


class RedisStorageData(RedisAdditionalData, types.MinimalStorageData):
    pass


class RedisUploader(Uploader):
    storage = None  # type: RedisStorage # pyright: ignore

    required_options = ["prefix"]
    capabilities = utils.combine_capabilities(
        Capability.CREATE,
        Capability.COPY,
        Capability.MOVE,
    )

    def upload(self, name, upload, extras):
        # type: (str, types.Upload, dict[str, types.Any]) -> RedisStorageData

        filename = self.storage.compute_name(name, extras, upload)
        key = self.storage.settings["prefix"] + filename

        self.storage.redis.delete(key)

        reader = HashingReader(upload.stream)
        self.storage.redis.set(key, b"")
        for chunk in reader:
            self.storage.redis.append(key, chunk)

        return {
            "filename": filename,
            "content_type": upload.content_type,
            "size": reader.position,
            "hash": reader.get_hash(),
        }


class RedisReader(Reader):
    storage = None  # type: RedisStorage # pyright: ignore

    required_options = ["prefix"]
    capabilities = utils.combine_capabilities(Capability.STREAM)

    def stream(self, data):
        # type: (dict[str, types.Any]) -> types.IO[bytes]
        return BytesIO(self.content(data))

    def content(self, data):
        # type: (dict[str, types.Any]) -> bytes
        key = self.storage.settings["prefix"] + data["filename"]
        value = self.storage.redis.get(key)
        if value is None:
            raise exceptions.MissingFileError(self.storage.settings["name"], key)

        return value


class RedisManager(Manager):
    storage = None  # type: RedisStorage # pyright: ignore

    required_options = ["prefix"]
    capabilities = utils.combine_capabilities(Capability.REMOVE, Capability.EXISTS)

    def remove(self, data):
        # type: (RedisStorageData) -> bool
        key = self.storage.settings["prefix"] + data["filename"]
        self.storage.redis.delete(key)
        return True

    def exists(self, data):
        # type: (RedisStorageData) -> bool
        key = self.storage.settings["prefix"] + data["filename"]
        return bool(self.storage.redis.exists(key))

    def copy(self, data, name, extras):
        # type: (types.MinimalStorageData, str, dict[str, types.Any]) -> RedisStorageData
        src = self.storage.settings["prefix"] + data["filename"]
        dest = self.storage.settings["prefix"] + name

        if not self.storage.redis.exists(src):
            raise exceptions.MissingFileError(self.storage.settings["name"], src)

        if self.storage.redis.exists(dest):
            raise exceptions.ExistingFileError(self.storage.settings["name"], dest)

        try:
            self.storage.redis.copy(src, dest)
        except AttributeError:
            self.storage.redis.restore(dest, 0, self.storage.redis.dump(src))

        return RedisStorageData(data, filename=name)

    def move(self, data, name, extras):
        # type: (types.MinimalStorageData, str, dict[str, types.Any]) -> RedisStorageData
        filename = self.storage.compute_name(name, extras)

        src = self.storage.settings["prefix"] + data["filename"]
        dest = self.storage.settings["prefix"] + filename

        self.storage.redis.rename(src, dest)
        return RedisStorageData(data, filename=filename)


class RedisStorage(Storage):
    def make_uploader(self):
        return RedisUploader(self)

    def make_manager(self):
        return RedisManager(self)

    def make_reader(self):
        return RedisReader(self)

    def __init__(self, **settings):
        # type: (**types.Any) -> None

        settings.setdefault(
            "prefix",
            _default_prefix(),
        )
        super(RedisStorage, self).__init__(**settings)
        self.redis = connect_to_redis()  # type: redis.Redis[bytes]

    @classmethod
    def declare_config_options(cls, declaration, key):
        # type: (types.Declaration, types.Key) -> None
        super().declare_config_options(declaration, key)
        declaration.declare(key.prefix, _default_prefix()).set_description(
            "Static prefix of the Redis key generated for every upload.",
        )


def _default_prefix():
    return "ckanext:files:{}:file_content:".format(tk.config["ckan.site_id"])



from .fs import FileSystemStorage, PublicFileSystemStorage
from .redis import RedisStorage

try:
    from .google_cloud import GoogleCloudStorage
except ImportError:  # pragma: no cover
    pass

__all__ = [
    "FileSystemStorage",
    "GoogleCloudStorage",
    "PublicFileSystemStorage",
    "RedisStorage",
]



"""Base abstract functionality of the extentsion.

All classes required for specific storage implementations are defined
here. Some utilities, like `storage_from_settings` are also added to this
module instead of `utils` to avoid import cycles.

This module relies only on types, exceptions and utils to prevent import
cycles.

"""

import abc
import copy
import hashlib
import os
import uuid
from datetime import datetime

from werkzeug.datastructures import FileStorage

from ckanext.files import exceptions, types, utils

CHUNK_SIZE = 16 * 1024
adapters = utils.Registry({})
storages = utils.Registry({})


def storage_from_settings(name, settings):
    # type: (str, dict[str, types.Any]) -> Storage
    """Initialize storage instance with specified settings.

    Storage adapter is defined by `type` key of the settings. All other
    settings depend on the specific adapter.

    Example:
    >>> storage = storage_from_settings("memo", {"type": "files:redis"})
    """

    adapter_type = settings.pop("type", None)
    adapter = adapters.get(adapter_type)  # type: type[Storage] | None
    if not adapter:
        raise exceptions.UnknownAdapterError(adapter_type)

    settings.setdefault("name", name)
    return adapter(**settings)


def get_storage(name):
    # type: (str) -> Storage
    """Return existing storage instance.

    Storages are initialized when plugin is loaded. As result, this function
    always returns the same storage object for the given name.

    Example:
    >>> storage = get_storage("default")
    """

    storage = storages.get(name)

    if not storage:
        raise exceptions.UnknownStorageError(name)

    return storage


class HashingReader:
    """IO stream wrapper that computes content hash while stream is consumed.

    Example:

    >>> reader = HashingReader(readable_stream)
    >>> for chunk in reader:
    >>>     ...
    >>> print(f"Hash: {reader.get_hash()}")

    """

    def __init__(self, stream, chunk_size=CHUNK_SIZE, algorithm="md5"):
        # type: (types.IO[bytes], int, str) -> None
        self.stream = stream
        self.chunk_size = chunk_size
        self.algorithm = algorithm
        self.hashsum = hashlib.new(algorithm)
        self.position = 0

    def __iter__(self):
        return self

    def __next__(self):
        chunk = self.stream.read(self.chunk_size)
        if not chunk:
            raise StopIteration

        self.position += len(chunk)
        self.hashsum.update(chunk)
        return chunk

    next = __next__

    def reset(self):
        """Rewind underlying stream and reset hash to initial state."""
        self.position = 0
        self.hashsum = hashlib.new(self.algorithm)
        self.stream.seek(0)

    def get_hash(self):
        """Get content hash as a string."""
        return self.hashsum.hexdigest()

    def exhaust(self):
        """Exhaust internal stream to compute final version of content hash."""

        for _ in self:
            pass


class Capability(object):
    """Enumeration of operations supported by the storage.

    Do not assume internal implementation of this type. Use Storage.supports,
    utils.combine_capabilities, and utils.exclude_capabilities to check and
    modify capabilities of the storage.

    Example:
    >>> read_and_write = utils.combine_capabilities(
    >>>     Capability.STREAM, Capability.CREATE,
    >>> )
    >>> storage.supports(read_and_write)
    """

    # create a file as an atomic object
    CREATE = types.CapabilityUnit(1 << 0)

    # return file content as stream of bytes
    STREAM = types.CapabilityUnit(1 << 1)

    # make a copy of the file inside the storage
    COPY = types.CapabilityUnit(1 << 2)

    # remove file from the storage
    REMOVE = types.CapabilityUnit(1 << 3)

    # create file in 3 stages: initialize, upload(repeatable), complete
    MULTIPART_UPLOAD = types.CapabilityUnit(1 << 4)

    # move file to a different location inside the storage
    MOVE = types.CapabilityUnit(1 << 5)

    # check if file exists
    EXISTS = types.CapabilityUnit(1 << 6)

    # iterate over all files in storage
    SCAN = types.CapabilityUnit(1 << 7)

    # add content to the existing file
    APPEND = types.CapabilityUnit(1 << 8)

    # combine multiple files into a new one
    COMPOSE = types.CapabilityUnit(1 << 9)

    # return specific range of file bytes
    RANGE = types.CapabilityUnit(1 << 10)

    # return file details from the storage
    ANALYZE = types.CapabilityUnit(1 << 11)


class OptionChecker(object):
    """Mixin for standard access to required settings.

    Example:
    >>> value = storage.ensure_option({...}, "option_name")

    """

    @classmethod
    def ensure_option(cls, settings, option):
        # type: (dict[str, types.Any], str) -> types.Any
        """Return value of the option or rise an exception."""

        if option not in settings:
            raise exceptions.MissingStorageConfigurationError(cls, option)
        return settings[option]


class StorageService(OptionChecker):
    """Base class for services used by storage.

    StorageService.capabilities reflect all operations provided by the
    service. StorageService.required_options list all options that are
    essential for the service. If any of this options is missing from the
    storage, service raises an error during storage initialization stage.

    >>> class Uploader(StorageService):
    >>>     capabilities = utils.combine_capabilities(Capability.CREATE)
    >>>     required_options = ["allowed_mimetypes"]

    """

    required_options = []  # type: list[str]
    capabilities = utils.combine_capabilities()

    def __init__(self, storage):
        # type: (Storage) -> None
        self.storage = storage
        self.ensure_settings()

    def ensure_settings(self):
        for option in self.required_options:
            self.ensure_option(self.storage.settings, option)


class Uploader(StorageService):
    """Service responsible for writing data into a storage."""

    def upload(self, name, upload, extras):
        # type: (str, types.Upload, dict[str, types.Any]) -> types.MinimalStorageData
        """Upload file using single stream."""

        raise NotImplementedError

    def initialize_multipart_upload(self, name, extras):
        # type: (str, dict[str, types.Any]) -> dict[str, types.Any]
        """Prepare everything for multipart(resumable) upload."""

        raise NotImplementedError

    def show_multipart_upload(self, upload_data):
        # type: (dict[str, types.Any]) -> dict[str, types.Any]
        """Show details of the incomplete upload."""
        raise NotImplementedError

    def update_multipart_upload(self, upload_data, extras):
        # type: (dict[str, types.Any], dict[str, types.Any]) -> dict[str, types.Any]
        """Add data to the incomplete upload."""
        raise NotImplementedError

    def complete_multipart_upload(self, upload_data, extras):
        # type: (dict[str, types.Any], dict[str, types.Any]) -> types.MinimalStorageData
        """Verify file integrity and finalize incomplete upload."""

        raise NotImplementedError


class Manager(StorageService):
    def remove(self, data):
        # type: (types.MinimalStorageData) -> bool
        """Remove file from the storage."""
        raise NotImplementedError

    def exists(self, data):
        # type: (types.MinimalStorageData) -> bool
        """Check if file exists in the storage."""

        raise NotImplementedError

    def copy(self, data, name, extras):
        # type: (types.MinimalStorageData, str, dict[str, types.Any]) -> types.MinimalStorageData
        """Copy file inside the storage."""

        raise NotImplementedError

    def move(self, data, name, extras):
        # type: (types.MinimalStorageData, str, dict[str, types.Any]) -> types.MinimalStorageData
        """Move file to a different location inside the storage."""
        raise NotImplementedError

    def scan(self):
        # type: () -> types.Iterable[str]
        """List all filenames in storage."""
        raise NotImplementedError

    def analyze(self, filename):
        # type: (str) -> types.MinimalStorageData
        """Return all details about filename."""
        raise NotImplementedError


class Reader(StorageService):
    def stream(self, data):
        # type: (types.MinimalStorageData) -> types.IO[bytes]
        """Return byte-stream of the file content."""

        raise NotImplementedError

    def content(self, data):
        # type: (types.MinimalStorageData) -> bytes
        """Return file content as a single byte object."""

        return self.stream(data).read()


class Storage(OptionChecker):
    __metaclass__ = abc.ABCMeta

    def __init__(self, **settings):
        # type: (**types.Any) -> None
        self.settings = settings

        self.uploader = self.make_uploader()
        self.manager = self.make_manager()
        self.reader = self.make_reader()

        self.capabilities = self.compute_capabilities()

    @property
    def max_size(self):
        size = self.settings.get("max_size", 0)

        # pre v2.10 CKAN instances do not support config declarations
        if isinstance(size, str):
            size = utils.parse_filesize(size)

        return size

    @classmethod
    def declare_config_options(cls, declaration, key):
        # type: (types.Declaration, types.Key) -> None
        declaration.declare(key.max_size, 0).append_validators(
            "files_parse_filesize",
        ).set_description(
            "The maximum size of a single upload."
            + "\nSupports size suffixes: 42B, 2M, 24KiB, 1GB."
            + " `0` means no restrictions.",
        )
        declaration.declare(key.name, key[-1]).set_description(
            "Descriptive name of the storage used for debugging.",
        )

    def compute_capabilities(self):
        # type: () -> types.CapabilityCluster
        return utils.combine_capabilities(
            self.uploader.capabilities,
            self.manager.capabilities,
            self.reader.capabilities,
        )

    def make_uploader(self):
        return Uploader(self)

    def make_manager(self):
        return Manager(self)

    def make_reader(self):
        return Reader(self)

    def supports(self, operation):
        # type: (types.CapabilityCluster | types.CapabilityUnit) -> bool
        return (self.capabilities & operation) == operation

    def compute_name(self, name, extras, upload=None):
        # type: (str, dict[str, types.Any], types.Upload|None) -> str
        strategy = self.settings.get("name_strategy", "uuid")
        if strategy == "uuid":
            return str(uuid.uuid4())

        if strategy == "uuid_prefix":
            return str(uuid.uuid4()) + name

        if strategy == "datetime_prefix":
            return datetime.utcnow().isoformat() + name

        if strategy == "uuid_with_extension":
            _path, ext = os.path.splitext(name)
            return str(uuid.uuid4()) + ext

        if strategy == "datetime_with_extension":
            _path, ext = os.path.splitext(name)
            return datetime.utcnow().isoformat() + ext

        raise exceptions.NameStrategyError(strategy)

    def upload(self, name, upload, extras):
        # type: (str, types.Upload, dict[str, types.Any]) -> types.MinimalStorageData
        if not self.supports(Capability.CREATE):
            raise exceptions.UnsupportedOperationError("upload", type(self).__name__)

        if self.max_size:
            utils.ensure_size(upload, self.max_size)

        return self.uploader.upload(name, upload, extras)

    def initialize_multipart_upload(self, name, extras):
        # type: (str, dict[str, types.Any]) -> dict[str, types.Any]
        return self.uploader.initialize_multipart_upload(name, extras)

    def show_multipart_upload(self, upload_data):
        # type: (dict[str, types.Any]) -> dict[str, types.Any]
        return self.uploader.show_multipart_upload(copy.deepcopy(upload_data))

    def update_multipart_upload(self, upload_data, extras):
        # type: (dict[str, types.Any], dict[str, types.Any]) -> dict[str, types.Any]
        return self.uploader.update_multipart_upload(copy.deepcopy(upload_data), extras)

    def complete_multipart_upload(self, upload_data, extras):
        # type: (dict[str, types.Any], dict[str, types.Any]) -> types.MinimalStorageData
        return self.uploader.complete_multipart_upload(
            copy.deepcopy(upload_data),
            extras,
        )

    def exists(self, data):
        # type: (types.MinimalStorageData) -> bool
        if not self.supports(Capability.EXISTS):
            raise exceptions.UnsupportedOperationError("exists", type(self).__name__)

        return self.manager.exists(data)

    def remove(self, data):
        # type: (types.MinimalStorageData) -> bool
        if not self.supports(Capability.REMOVE):
            raise exceptions.UnsupportedOperationError("remove", type(self).__name__)

        return self.manager.remove(data)

    def scan(self):
        # type: () -> types.Iterable[str]
        if not self.supports(Capability.SCAN):
            raise exceptions.UnsupportedOperationError("scan", type(self).__name__)

        return self.manager.scan()

    def analyze(self, filename):
        # type: (str) -> types.MinimalStorageData
        if not self.supports(Capability.ANALYZE):
            raise exceptions.UnsupportedOperationError("analyze", type(self).__name__)

        return self.manager.analyze(filename)

    def stream(self, data):
        # type: (types.MinimalStorageData) -> types.IO[bytes]
        if not self.supports(Capability.STREAM):
            raise exceptions.UnsupportedOperationError("stream", type(self).__name__)

        return self.reader.stream(data)

    def content(self, data):
        # type: (types.MinimalStorageData) -> bytes | str
        if not self.supports(Capability.STREAM):
            raise exceptions.UnsupportedOperationError("content", type(self).__name__)

        return self.reader.content(data)

    def copy(self, data, storage, name, extras):
        # type: (types.MinimalStorageData, Storage, str, dict[str, types.Any]) -> types.MinimalStorageData
        if storage is self and self.supports(Capability.COPY):
            return self.manager.copy(data, name, extras)

        if self.supports(Capability.STREAM) and storage.supports(Capability.CREATE):
            return storage.upload(name, FileStorage(self.stream(data)), extras)

        raise exceptions.UnsupportedOperationError("copy", type(self).__name__)

    def move(self, data, storage, name, extras):
        # type: (types.MinimalStorageData, Storage, str, dict[str, types.Any]) -> types.MinimalStorageData

        if storage is self and self.supports(Capability.MOVE):
            return self.manager.move(data, name, extras)

        if self.supports(
            utils.combine_capabilities(Capability.STREAM, Capability.REMOVE),
        ) and storage.supports(Capability.CREATE):
            result = storage.upload(name, FileStorage(self.stream(data)), extras)
            storage.remove(data)
            return result

        raise exceptions.UnsupportedOperationError("copy", type(self).__name__)



import hashlib
import os
from io import BytesIO
from uuid import UUID

import pytest
from werkzeug.datastructures import FileStorage

import ckan.plugins.toolkit as tk

from ckanext.files import exceptions
from ckanext.files.storage import fs

from faker import Faker  # isort: skip # noqa: F401


@pytest.fixture()
def storage(clean_redis, tmp_path):
    # type: (object, object) -> fs.FileSystemStorage
    return fs.FileSystemStorage(name="test", path=str(tmp_path))


class TestUploader:
    def test_key(self, storage):
        # type: (fs.FileSystemStorage) -> None
        result = storage.upload("", FileStorage(), {})

        assert UUID(result["filename"])
        assert result["size"] == 0

        filepath = os.path.join(storage.settings["path"], result["filename"])

        assert os.path.exists(filepath)
        assert os.path.getsize(filepath) == 0

    def test_content(self, storage, faker):
        # type: (fs.FileSystemStorage, Faker) -> None
        content = faker.binary(100)
        result = storage.upload("", FileStorage(BytesIO(content)), {})

        assert result["size"] == 100

        filepath = os.path.join(storage.settings["path"], result["filename"])
        assert open(filepath, "rb").read() == content

    def test_hash(self, storage, faker):
        # type: (fs.FileSystemStorage, Faker) -> None
        result = storage.upload("", FileStorage(), {})
        assert result["hash"] == hashlib.md5().hexdigest()

        content = faker.binary(100)
        result = storage.upload("", FileStorage(BytesIO(content)), {})
        assert result["hash"] == hashlib.md5(content).hexdigest()


@pytest.mark.usefixtures("with_plugins")
class TestMultipartUploader:
    def test_initialization_invalid(self, storage, faker):
        # type: (fs.FileSystemStorage, Faker) -> None
        with pytest.raises(tk.ValidationError):
            storage.initialize_multipart_upload(faker.file_name(), {})

    def test_initialization_large(self, storage, faker):
        # type: (fs.FileSystemStorage, Faker) -> None
        storage.settings["max_size"] = 5
        with pytest.raises(exceptions.LargeUploadError):
            storage.initialize_multipart_upload(faker.file_name(), {"size": 10})

    def test_initialization(self, storage, faker):
        # type: (fs.FileSystemStorage, Faker) -> None
        content = b"hello world"
        data = storage.initialize_multipart_upload(
            faker.file_name(),
            {"size": len(content)},
        )
        assert data["size"] == len(content)
        assert data["uploaded"] == 0

    def test_update_invalid(self, storage, faker):
        # type: (fs.FileSystemStorage, Faker) -> None
        content = b"hello world"
        data = storage.initialize_multipart_upload(
            faker.file_name(),
            {"size": len(content)},
        )
        with pytest.raises(tk.ValidationError):
            storage.update_multipart_upload(data, {})

    def test_update(self, storage, faker):
        # type: (fs.FileSystemStorage, Faker) -> None
        content = b"hello world"
        data = storage.initialize_multipart_upload(
            faker.file_name(),
            {"size": len(content)},
        )

        data = storage.update_multipart_upload(
            data,
            {"upload": FileStorage(BytesIO(content[:5]))},
        )
        assert data["size"] == len(content)
        assert data["uploaded"] == 5

        data = storage.update_multipart_upload(
            data,
            {"upload": FileStorage(BytesIO(content[:5])), "position": 3},
        )
        assert data["size"] == len(content)
        assert data["uploaded"] == 8

        with pytest.raises(exceptions.UploadOutOfBoundError):
            storage.update_multipart_upload(
                data,
                {"upload": FileStorage(BytesIO(content))},
            )

        missing_size = data["size"] - data["uploaded"]
        data = storage.update_multipart_upload(
            data,
            {"upload": FileStorage(BytesIO(content[-missing_size:]))},
        )
        assert data["size"] == len(content)
        assert data["uploaded"] == len(content)

    def test_complete(self, storage, faker):
        # type: (fs.FileSystemStorage, Faker) -> None
        content = b"hello world"
        data = storage.initialize_multipart_upload(
            faker.file_name(),
            {"size": len(content)},
        )

        with pytest.raises(tk.ValidationError):
            storage.complete_multipart_upload(data, {})

        data = storage.update_multipart_upload(
            data,
            {"upload": FileStorage(BytesIO(content))},
        )
        data = storage.complete_multipart_upload(data, {})
        assert data["size"] == len(content)
        assert data["hash"] == hashlib.md5(content).hexdigest()

    def test_show(self, storage, faker):
        # type: (fs.FileSystemStorage, Faker) -> None
        content = b"hello world"

        data = storage.initialize_multipart_upload(
            faker.file_name(),
            {"size": len(content)},
        )
        assert storage.show_multipart_upload(data) == data

        data = storage.update_multipart_upload(
            data,
            {"upload": FileStorage(BytesIO(content))},
        )
        assert storage.show_multipart_upload(data) == data

        storage.complete_multipart_upload(data, {})
        assert storage.show_multipart_upload(data) == data


class TestManager:
    def test_removal(self, storage):
        # type: (fs.FileSystemStorage) -> None
        result = storage.upload("", FileStorage(), {})
        filepath = os.path.join(storage.settings["path"], result["filename"])
        assert os.path.exists(filepath)

        assert storage.remove(result)
        assert not os.path.exists(filepath)

    def test_removal_missing(self, storage):
        # type: (fs.FileSystemStorage) -> None
        result = storage.upload("", FileStorage(), {})
        assert storage.remove(result)
        assert not storage.remove(result)


class TestReader:
    def test_stream(self, storage, faker):
        # type: (fs.FileSystemStorage, Faker) -> None
        data = faker.binary(100)
        result = storage.upload("", FileStorage(BytesIO(data)), {})

        stream = storage.stream(result)

        assert stream.read() == data

    def test_content(self, storage, faker):
        # type: (fs.FileSystemStorage, Faker) -> None
        data = faker.binary(100)
        result = storage.upload("", FileStorage(BytesIO(data)), {})

        content = storage.content(result)

        assert content == data

    def test_missing(self, storage, faker):
        # type: (fs.FileSystemStorage, Faker) -> None
        result = storage.upload("", FileStorage(), {})
        result["filename"] += faker.uuid4()

        with pytest.raises(exceptions.MissingFileError):
            storage.stream(result)

        with pytest.raises(exceptions.MissingFileError):
            storage.content(result)


class TestStorage:
    def test_missing_path(self, tmp_path):
        # type: (str) -> None
        with pytest.raises(exceptions.InvalidStorageConfigurationError):
            fs.FileSystemStorage(path=os.path.join(str(tmp_path), "not-real"))

    def test_missing_path_created(self, tmp_path):
        # type: (str) -> None
        path = os.path.join(str(tmp_path), "not-real")
        assert not os.path.exists(path)

        fs.FileSystemStorage(path=path, create_path=True)
        assert os.path.exists(path)



import glob
import os

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan import model
from ckan.lib import uploader

from ckanext.files.model import File
from ckanext.files.storage import fs

from ckanext.files import types, interfaces, shared  # isort: skip # noqa: F401


class FilesUploaderPlugin(p.SingletonPlugin):
    p.implements(p.IUploader)
    p.implements(interfaces.IFiles)

    def files_get_storage_adapters(self):
        # type: () -> dict[str, types.Any]

        return {
            "files_uploader:resource": ResourceStorage,
            "files_uploader:image": ImageStorage,
        }

    # IUploader
    def get_uploader(self, upload_to, old_filename):
        # type: (str, str | None) -> types.PUploader | None
        return Uploader(upload_to, old_filename)

    def get_resource_uploader(self, resource):
        # type: (dict[str, types.Any]) -> types.PResourceUploader | None
        return ResourceUploader(resource)


class Uploader(uploader.Upload):
    def __init__(self, object_type, old_filename=None):
        # type: (str, str | None) -> None
        self.storage = shared.get_storage(image_storage_name())
        super(Uploader, self).__init__(object_type, old_filename)

    def upload(self, max_size=2):
        # type: (int) -> None
        if hasattr(self, "verify_type"):
            self.verify_type()

        if self.filename:
            tk.get_action("files_file_create")(
                {"ignore_auth": True},
                {
                    "name": self.filename,
                    "upload": self.upload_file,
                    "storage": self.storage.settings["name"],
                    "object_type": self.object_type,
                },
            )
            self.clear = True

        if (
            self.clear
            and self.old_filename
            and not self.old_filename.startswith("http")
            and self.old_filepath
        ):
            existing = model.Session.execute(
                File.by_location(self.old_filename, self.storage.settings["name"]),
            ).scalar()
            if not existing:
                return
            tk.get_action("files_file_delete")(
                {"ignore_auth": True},
                {"id": existing.id},
            )


class ResourceUploader(uploader.ResourceUpload):
    # mimetype: Optional[str]
    # filesize: int

    def __init__(self, resource):
        # type: (dict[str, types.Any]) -> None
        self.storage = shared.get_storage(resource_storage_name())
        super(ResourceUploader, self).__init__(resource)

    def get_path(self, id):
        # type: (str) -> str
        return os.path.join(
            self.storage.settings["path"],
            self._file_location(id),
        )

    def _file_location(self, id):
        # type: (str) -> str
        return self.storage.compute_name(id, {"resource_id": id}, None)

    def upload(self, id, max_size=10):
        # type: (str, int) -> None
        # If a filename has been provided (a file is being uploaded)
        # we write it to the filepath (and overwrite it if it already
        # exists). This way the uploaded file will always be stored
        # in the same location
        if self.filename and self.upload_file:
            tk.get_action("files_file_create")(
                {"ignore_auth": True},
                {
                    "name": self.filename,
                    "upload": self.upload_file,
                    "storage": self.storage.settings["name"],
                    "resource_id": id,
                },
            )

        # The resource form only sets self.clear (via the input clear_upload)
        # to True when an uploaded file is not replaced by another uploaded
        # file, only if it is replaced by a link to file.
        # If the uploaded file is replaced by a link, we should remove the
        # previously uploaded file to clean up the file system.
        if self.clear:
            existing = model.Session.execute(
                File.by_location(
                    self._file_location(id),
                    self.storage.settings["name"],
                ),
            ).scalar()
            if not existing:
                return
            tk.get_action("files_file_delete")(
                {"ignore_auth": True},
                {"id": existing.id},
            )


class ImageStorage(fs.FileSystemStorage):
    def compute_name(self, name, extras, upload=None):
        # type: (str, dict[str, types.Any], types.Upload|None) -> str
        return os.path.join("storage", "uploads", extras["object_type"], name)

    def make_manager(self):
        return ImageManager(self)


class ImageManager(fs.FileSystemManager):
    capabilities = shared.combine_capabilities(
        fs.FileSystemManager.capabilities,
        shared.Capability.SCAN,
    )

    def scan(self):
        # type: () -> types.Iterable[str]
        path = self.storage.settings["path"]
        for name in glob.iglob(os.path.join(path, "uploads", "*", "*")):
            yield os.path.relpath(name, path)


class ResourceStorage(fs.FileSystemStorage):
    def compute_name(self, name, extras, upload=None):
        # type: (str, dict[str, types.Any], types.Upload|None) -> str
        res_id = extras["resource_id"]
        return os.path.join("resources", res_id[0:3], res_id[3:6], res_id[6:])

    def make_manager(self):
        return ResourceManager(self)


class ResourceManager(fs.FileSystemManager):
    capabilities = shared.combine_capabilities(
        fs.FileSystemManager.capabilities,
        shared.Capability.SCAN,
    )

    def scan(self):
        # type: () -> types.Iterable[str]
        path = self.storage.settings["path"]
        for name in glob.iglob(os.path.join(path, "resources", "???", "???", "*")):
            yield os.path.relpath(name, path)


def resource_storage_name():
    # type: () -> str
    return tk.config.get("ckanext.files_uploader.resource_storage", "resource")


def image_storage_name():
    # type: () -> str
    return tk.config.get("ckanext.files_uploader.image_storage", "image")



"""Create files table

Revision ID: cc1a832108c5
Revises:
Create Date: 2021-09-21 13:02:25.731642

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import JSONB

# revision identifiers, used by Alembic.
revision = "cc1a832108c5"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "files_file",
        sa.Column("id", sa.UnicodeText, primary_key=True),
        sa.Column("name", sa.UnicodeText, nullable=False),
        sa.Column("path", sa.UnicodeText, nullable=False),
        sa.Column("kind", sa.UnicodeText, nullable=False),
        sa.Column(
            "uploaded_at",
            sa.DateTime,
            nullable=False,
            server_default=sa.func.current_timestamp(),
        ),
        sa.Column("extras", JSONB),
    )


def downgrade():
    op.drop_table("files_file")



import pytest

import ckan.model as model
from ckan.tests.helpers import call_action

from ckanext.files.model import File

from ckanext.files import types  # isort: skip # noqa: F401


@pytest.fixture
def random_file(create_with_upload, faker):
    return create_with_upload(
        faker.binary(10),
        faker.file_name(),
        action="files_file_create",
        name=faker.name(),
    )


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileCreate:
    def test_basic_file(self, create_with_upload):
        filename = "file.txt"
        result = create_with_upload(
            "hello",
            filename,
            action="files_file_create",
            name="Test file.txt",
        )

        assert result["name"] == "Test file.txt"


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileDelete:
    def test_basic_delete(self, random_file):
        # type: (dict[str, types.Any]) -> None
        q = model.Session.query(File)
        assert q.count() == 1
        call_action("files_file_delete", id=random_file["id"])
        assert not q.count()


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileShow:
    def test_basic_show(self, random_file):
        # type: (dict[str, types.Any]) -> None
        result = call_action("files_file_show", id=random_file["id"])
        assert result["id"] == random_file["id"]

        result = call_action("files_file_show", id=random_file["id"])
        assert result["id"] == random_file["id"]



"""Types for the extension.

These bizarre definitions must remain here as long as Python2 is supported.

"""

import six
from sqlalchemy.sql.expression import Select
from werkzeug.datastructures import FileStorage as Upload

import ckan.plugins.toolkit as tk

CapabilityCluster = int
CapabilityUnit = int
TypedDict = lambda _name, _fields, **kwargs: type(_name, (dict,), {})  # type: ignore # noqa: E731
Declaration = None
Key = None
Any = None
IO = None
Iterable = None
PUploader = None
PResourceUploader = None
cast = lambda _type, _value: _value  # type: ignore # noqa: E731

if six.PY3:
    from typing import (  # pyright: ignore[reportConstantRedefinition]
        IO,
        Any,
        Iterable,
        cast,
    )

    from typing_extensions import NewType, TypedDict

    CapabilityCluster = NewType("CapabilityCluster", int)
    CapabilityUnit = NewType("CapabilityUnit", int)

    if tk.check_ckan_version("2.10"):
        from ckan.config.declaration import Declaration, Key
        from ckan.types import PResourceUploader, PUploader

__all__ = [
    "Upload",
    "CapabilityCluster",
    "CapabilityUnit",
    "TypedDict",
    "MinimalStorageData",
    "Key",
    "Declaration",
    "Any",
    "IO",
    "Select",
    "PUploader",
    "PResourceUploader",
    "cast",
    "Iterable",
]


MinimalStorageData = TypedDict(
    "MinimalStorageData",
    {"filename": str, "content_type": str, "size": int, "hash": str},
)



import six
import sqlalchemy as sa

import ckan.plugins.toolkit as tk
from ckan.lib.dictization import table_dictize
from ckan.model.types import make_uuid

from .base import Base

from ckanext.files import types  # isort: skip # noqa: F401

if six.PY3:
    from typing import Any  # isort: skip # noqa: F401


class Owner(Base):  # type: ignore
    __tablename__ = "files_owner"
    id = sa.Column(sa.Text, primary_key=True, default=make_uuid)

    item_id = sa.Column(sa.Text, nullable=False)
    item_type = sa.Column(sa.Text, nullable=False)
    owner_id = sa.Column(sa.Text, nullable=False)
    owner_type = sa.Column(sa.Text, nullable=False)
    access = sa.Column(sa.Text, nullable=False, default="full")

    sa.UniqueConstraint(item_id, item_type, owner_id, owner_type)

    def dictize(self, context):
        # type: (Any) -> dict[str, Any]
        return table_dictize(self, context)

    @classmethod
    def owners_of(cls, id, type):
        # type: (str, str) -> types.Select
        """List owners of the given item."""
        selectable = cls if tk.check_ckan_version("2.9") else [cls]
        return sa.select(selectable).where(
            sa.and_(cls.item_type == type, cls.item_id == id),
        )

    @classmethod
    def owned_by(cls, id, type):
        # type: (str, str) -> types.Select
        """List records with given owner."""
        selectable = cls if tk.check_ckan_version("2.9") else [cls]
        return sa.select(selectable).where(
            sa.and_(cls.owner_type == type, cls.owner_id == id),
        )



import sqlalchemy as sa

import ckan.plugins.toolkit as tk
from ckan import authz, model

from ckanext.files.model import Owner
from ckanext.files.utils import make_collector

from ckanext.files import types  # isort: skip # noqa: F401


_auth_functions, auth = make_collector()


def get_auth_functions():
    return dict(_auth_functions)


def _get_user(context):
    # type: (types.Any) -> model.User | None
    if "auth_user_obj" in context:
        return context["auth_user_obj"]

    if tk.check_ckan_version("2.10"):
        user = tk.current_user if tk.current_user.is_authenticated else None
    else:
        user = tk.g.userobj  # type: types.Any

    if user and context["user"] == user.name:
        return user

    return model.User.get(context["user"])


def _is_owner(user_id, file_id):
    # type: (str, str) -> bool
    stmt = Owner.owners_of(file_id, "file").where(
        sa.and_(
            Owner.owner_type == "user",
            Owner.owner_ie == user_id,
        ),
    )
    return model.Session.query(stmt.exists()).scalar()


@auth
@tk.auth_disallow_anonymous_access
def files_manage_files(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> types.Any
    return {"success": False}


@auth
@tk.auth_disallow_anonymous_access
def files_owns_file(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> types.Any
    user = _get_user(context)
    is_owner = bool(user and _is_owner(user.id, data_dict["id"]))

    return {"success": is_owner, "msg": "Not an owner of the file"}


@auth
@tk.auth_disallow_anonymous_access
def files_file_search_by_user(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> types.Any
    """Only user himself can view his own files."""

    # `user` from context will be used used when it's not in data_dict, so it's
    # an access to own files
    if "user" not in data_dict:
        return {"success": True}

    user = _get_user(context)

    return {
        "success": user and data_dict["user"] in [user.name, user.id],
        "msg": "Not authorized to view files of this user",
    }


@auth
@tk.auth_disallow_anonymous_access
def files_file_create(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> types.Any
    return authz.is_authorized("files_manage_files", context, data_dict)


@auth
@tk.auth_disallow_anonymous_access
def files_file_delete(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> types.Any
    """Only owner can remove files."""
    return authz.is_authorized("files_owns_file", context, data_dict)


@auth
@tk.auth_disallow_anonymous_access
def files_file_show(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> types.Any
    """Only owner can view files."""
    return authz.is_authorized("files_owns_file", context, data_dict)


@auth
@tk.auth_disallow_anonymous_access
def files_file_rename(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> types.Any
    """Only owner can rename files."""
    return authz.is_authorized("files_owns_file", context, data_dict)


@auth
@tk.auth_disallow_anonymous_access
def files_upload_show(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> types.Any
    return authz.is_authorized("files_owns_file", context, data_dict)


@auth
@tk.auth_disallow_anonymous_access
def files_upload_initialize(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> types.Any
    return authz.is_authorized("files_file_create", context, data_dict)


@auth
@tk.auth_disallow_anonymous_access
def files_upload_update(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> types.Any
    return authz.is_authorized("files_owns_file", context, data_dict)


@auth
@tk.auth_disallow_anonymous_access
def files_upload_complete(context, data_dict):
    # type: (types.Any, dict[str, types.Any]) -> types.Any
    return authz.is_authorized("files_owns_file", context, data_dict)



from ckanext.files.base import (
    Capability,
    Manager,
    Reader,
    Storage,
    Uploader,
    get_storage,
)
from ckanext.files.utils import combine_capabilities, exclude_capabilities

__all__ = [
    "combine_capabilities",
    "exclude_capabilities",
    "get_storage",
    "Storage",
    "Uploader",
    "Reader",
    "Manager",
    "Capability",
]



"""Configuration readers of the extension.

This module contains functions that simplify accessing configuration option
from the CKAN config file.

It's recommended to use these functions istead of accessing config options by
name, if you want your code to be more compatible with different versions of
the extension.

"""

from collections import defaultdict

import six

import ckan.plugins.toolkit as tk

if six.PY3:
    from typing import Any  # isort: skip # noqa: F401


DEFAULT_STORAGE = "ckanext.files.default_storage"
STORAGE_PREFIX = "ckanext.files.storage."


def default_storage():
    # type: () -> str
    """Default storage used for upload when no explicit storage specified."""

    return tk.config.get(DEFAULT_STORAGE, "default")


def storages():
    # type: () -> dict[str, dict[str, Any]]
    """Mapping of storage names to their settings."""
    storages = defaultdict(dict)  # type: dict[str, dict[str, Any]]
    prefix_len = len(STORAGE_PREFIX)
    for k, v in tk.config.items():
        if not k.startswith(STORAGE_PREFIX):
            continue

        try:
            name, option = k[prefix_len:].split(".", 1)
        except ValueError:
            continue

        storages[name][option] = v
    return storages



"""Internal utilities of the extension.

Do not use this module outside of the extension and do not import any other
internal modules except for types and exceptions. Only independent tools are
stored here, to avoid import cycles.

"""

import cgi
import mimetypes
import re
import tempfile
from io import BytesIO

import magic
import six
from werkzeug.datastructures import FileStorage

from ckanext.files import exceptions

from ckanext.files import types  # isort: skip # noqa: 401

if six.PY3:
    from typing import TypeVar

    from typing import Callable, Any  # isort: skip # noqa: F401

    T = TypeVar("T")


RE_FILESIZE = re.compile(r"^(?P<size>\d+(?:\.\d+)?)\s*(?P<unit>\w*)$")

UNITS = {
    "": 1,
    "b": 1,
    "k": 10**3,
    "kb": 10**3,
    "m": 10**6,
    "mb": 10**6,
    "g": 10**9,
    "gb": 10**9,
    "t": 10**12,
    "tb": 10**12,
    "kib": 2**10,
    "mib": 2**20,
    "gib": 2**30,
    "tib": 2**40,
}  # type: dict[str, int]


class Registry(object):
    """Mutable collection of objects.

    Example:
    >>> col = Registry()
    >>>
    >>> col.register("one", 1)
    >>> assert col.get("one") == 1
    >>>
    >>> col.reset()
    >>> assert col.get("one") is None
    """

    def __init__(self, members=None):
        # type: (dict[str, Any] | None) -> None
        if members is None:
            members = {}
        self.members = members

    def __iter__(self):
        return iter(self.members)

    def reset(self):
        """Remove all members from registry."""

        self.members.clear()

    def register(self, name, member):
        # type: (str, Any) -> None
        """Add a member to registry."""

        self.members[name] = member

    def get(self, name):
        # type: (Any) -> Any | None
        """Get an optional member from registry."""

        return self.members.get(name)


def make_collector():
    # type: () -> tuple[dict[str, Any], Callable[[Any], Any]]
    """Create pair of a dictionary and decorator that appends function to the
    dictionary.

    Example:
    >>> col, add = make_collector()
    >>> assert col == {}
    >>>
    >>> @add
    >>> def hello():
    >>>     return "world"
    >>>
    >>> assert col == {"hello": hello}
    """

    collection = {}  # type: dict[str, Any]

    def collector(fn):
        # type: (T) -> T
        """Decorator that appends functions to the collection."""

        collection[fn.__name__] = fn
        return fn

    return collection, collector


def ensure_size(upload, max_size):
    # type: (types.Upload, int) -> int
    """Return filesize or rise an exception if it exceedes max_size."""

    filesize = upload.content_length
    if not filesize:
        upload.stream.seek(0, 2)
        filesize = upload.stream.tell()
        upload.stream.seek(0)

    if filesize > max_size:
        raise exceptions.LargeUploadError(filesize, max_size)

    return filesize


def combine_capabilities(*capabilities):
    # type: (*types.CapabilityCluster | types.CapabilityUnit) -> types.CapabilityCluster
    """Combine multiple capabilities.

    Example:
    >>> cluster = combine_capabilities(Capability.CREATE, Capability.REMOVE)
    """

    result = 0
    for capability in capabilities:
        result |= capability

    return types.CapabilityCluster(result)


def exclude_capabilities(capabilities, *exclude):
    # type: (types.CapabilityCluster, *types.CapabilityCluster | types.CapabilityUnit) -> types.CapabilityCluster
    """Remove capabilities from the cluster

    Example:
    >>> cluster = exclude_capabilities(cluster, Capability.REMOVE)
    """

    for capability in exclude:
        capabilities = types.CapabilityCluster(capabilities & ~capability)

    return capabilities


def parse_filesize(value):
    # type: (str) -> int
    """Transform human-readable filesize into an integer.

    Example:
    >>> size = parse_filesize("10GiB")
    >>> assert size == 10_737_418_240
    """
    result = RE_FILESIZE.match(value.strip())
    if not result:
        raise ValueError(value)
    size, unit = result.groups()

    multiplier = UNITS.get(unit.lower())
    if not multiplier:
        raise ValueError(value)

    return int(float(size) * multiplier)


def make_upload(value):
    # type: (Any) -> types.Upload
    """Convert value into werkzeug.FileStorage object"""
    if isinstance(value, FileStorage):
        if not value.content_length:
            value.stream.seek(0, 2)
            value.headers["content-length"] = str(value.stream.tell())
            value.stream.seek(0)
        return value

    if isinstance(value, tempfile.SpooledTemporaryFile):
        mime = magic.from_buffer(value.read(1024), True)
        value.seek(0, 2)
        size = value.tell()
        value.seek(0)

        return FileStorage(value, "", content_type=mime, content_length=size)

    if isinstance(value, cgi.FieldStorage):
        if not value.filename or not value.file:
            raise ValueError(value)

        mime, _encoding = mimetypes.guess_type(value.filename)
        if not mime:
            mime = magic.from_buffer(value.file.read(1024), True)
            value.file.seek(0)
        value.file.seek(0, 2)
        size = value.file.tell()
        value.file.seek(0)

        return FileStorage(
            value.file,
            value.filename,
            content_type=mime,
            content_length=size,
        )

    if isinstance(value, six.text_type):
        value = value.encode()

    if isinstance(value, (bytes, bytearray)):
        value = BytesIO(value)

    if isinstance(value, BytesIO):
        mime = magic.from_buffer(value.read(1024), True)
        value.seek(0, 2)
        size = value.tell()
        value.seek(0)

        return FileStorage(value, content_type=mime, content_length=size)

    raise TypeError(type(value))



from io import BytesIO

import pytest
from werkzeug.datastructures import FileStorage

from ckanext.files import base, exceptions, utils

from faker import Faker  # isort: skip # noqa: F401


def test_registry(faker):
    # type: (Faker) -> None
    """Brief test of registry functionality."""

    registry = utils.Registry()
    key = faker.word()
    value = object()

    assert registry.get(key) is None

    registry.register(key, value)
    assert registry.get(key) is value
    assert list(registry) == [key]

    registry.reset()
    assert registry.get(key) is None
    assert list(registry) == []


def test_collector():
    """Brief test of collector functionality."""

    collection, collector = utils.make_collector()
    assert collection == {}

    @collector
    def func():  # pragma: no cover
        pass

    assert collection == {"func": func}


class TestEnsureSize:
    def test_empty(self):
        """Filesize identified even if it's not set initially."""

        assert utils.ensure_size(FileStorage(), 0) == 0

    def test_not_empty(self):
        """Big files cause exception."""

        upload = FileStorage(BytesIO(b" " * 10))
        assert utils.ensure_size(upload, 15) == 10

        with pytest.raises(exceptions.LargeUploadError):
            utils.ensure_size(upload, 5)


class TestCombineCapabilities:
    def test_reflexive_combination(self):
        """Combination of unit with itself leaves a single unit."""

        first = utils.combine_capabilities(
            base.Capability.CREATE,
            base.Capability.CREATE,
        )
        second = utils.combine_capabilities(base.Capability.CREATE)
        assert first == second

    def test_commutative_combination(self):
        """Order of combination does not change the result"""

        first = utils.combine_capabilities(
            base.Capability.CREATE,
            base.Capability.REMOVE,
        )
        second = utils.combine_capabilities(
            base.Capability.REMOVE,
            base.Capability.CREATE,
        )
        assert first == second

    def test_associative_combination(self):
        """Rearranging the combination sequence does not change the result."""

        first = utils.combine_capabilities(
            utils.combine_capabilities(base.Capability.CREATE, base.Capability.REMOVE),
            base.Capability.MULTIPART_UPLOAD,
        )
        second = utils.combine_capabilities(
            utils.combine_capabilities(
                base.Capability.CREATE,
                base.Capability.MULTIPART_UPLOAD,
            ),
            base.Capability.REMOVE,
        )
        third = utils.combine_capabilities(
            utils.combine_capabilities(
                base.Capability.REMOVE,
                base.Capability.MULTIPART_UPLOAD,
            ),
            base.Capability.CREATE,
        )

        assert first == second == third

    def test_clusters(self):
        """Clusters can be combined with the same result as individual units."""

        first = utils.combine_capabilities(
            base.Capability.CREATE,
            base.Capability.REMOVE,
        )
        second = utils.combine_capabilities(
            base.Capability.MULTIPART_UPLOAD,
            base.Capability.STREAM,
        )

        clusters = utils.combine_capabilities(first, second)
        units = utils.combine_capabilities(
            base.Capability.CREATE,
            base.Capability.REMOVE,
            base.Capability.MULTIPART_UPLOAD,
            base.Capability.STREAM,
        )
        assert clusters == units


class TestExcludeCapabilities:
    def test_not_intersecting_exclusion(self):
        """Nothing changes when non existing unit excluded."""

        cluster = utils.combine_capabilities(
            base.Capability.CREATE,
            base.Capability.REMOVE,
        )

        assert (
            utils.exclude_capabilities(cluster, base.Capability.MULTIPART_UPLOAD)
            == cluster
        )

    def test_exclusion_of_single_unit(self):
        """Single unit exclusion leaves all other units inside cluster."""

        cluster = utils.combine_capabilities(
            base.Capability.CREATE,
            base.Capability.REMOVE,
        )

        assert utils.exclude_capabilities(
            cluster,
            base.Capability.CREATE,
        ) == utils.combine_capabilities(base.Capability.REMOVE)

    def test_multi_unit_exclusion(self):
        """Multiple units can be excluded at once."""

        cluster = utils.combine_capabilities(
            base.Capability.CREATE,
            base.Capability.REMOVE,
            base.Capability.STREAM,
        )
        assert utils.exclude_capabilities(
            cluster,
            base.Capability.REMOVE,
            base.Capability.CREATE,
        ) == utils.combine_capabilities(base.Capability.STREAM)

    def test_exclusion_of_cluster(self):
        """The whole cluster can be excluded at once."""

        cluster = utils.combine_capabilities(
            base.Capability.CREATE,
            base.Capability.REMOVE,
            base.Capability.STREAM,
        )

        empty = utils.exclude_capabilities(
            cluster,
            utils.combine_capabilities(base.Capability.CREATE, base.Capability.STREAM),
        )
        assert empty == utils.combine_capabilities(base.Capability.REMOVE)


class TestParseFilesize:
    @pytest.mark.parametrize(
        ["value", "size"],
        [
            ("1", 1),
            ("1b", 1),
            ("1kb", 10**3),
            ("1kib", 2**10),
            ("1mb", 10**6),
            ("1mib", 2**20),
            ("1gb", 10**9),
            ("1gib", 2**30),
            ("1tb", 10**12),
            ("1tib", 2**40),
            ("  117  ", 117),
            ("0.7 mib", 734003),
            ("1b", 1),
            ("1 kib", 1024),
            ("10.43 kib", 10680),
            ("1024b", 1024),
            ("11 GiB", 11811160064),
            ("117 b", 117),
            ("117 kib", 119808),
            ("117b", 117),
            ("117kib", 119808),
            ("11GiB", 11811160064),
            ("1mib", 1048576),
            ("343.1MiB", 359766425),
            ("5.2 mib", 5452595),
            ("58 kib", 59392),
        ],
    )
    def test_valid_sizes(self, value, size):
        # type: (str, int) -> None
        """Human-readable filesize is parsed into number of bytes."""

        assert utils.parse_filesize(value) == size

    def test_empty_string(self):
        """Empty string causes an exception"""
        with pytest.raises(ValueError):
            utils.parse_filesize("")

    def test_invalid_multiplier(self):
        """Empty string causes an exception"""

        with pytest.raises(ValueError):
            utils.parse_filesize("1PB")



from __future__ import annotations

import logging
from typing import Callable, Union

from flask import Blueprint, Response
from flask.views import MethodView

import ckan.plugins.toolkit as tk
from ckan.logic import parse_params

from ckanext.ap_main.utils import ap_before_request
from ckanext.collection.shared import get_collection

log = logging.getLogger(__name__)
file_manager = Blueprint(
    "file_manager",
    __name__,
    url_prefix="/admin-panel/file_manager",
)
file_manager.before_request(ap_before_request)


class FileManagerView(MethodView):
    def get(self) -> Union[str, Response]:
        return tk.render(
            "file_manager/list.html",
            extra_vars={
                "collection": get_collection(
                    "file-manager",
                    parse_params(tk.request.args),
                ),
            },
        )

    def post(self) -> Response:
        bulk_action = tk.request.form.get("bulk-action")
        file_ids = tk.request.form.getlist("entity_id")

        action_func = self._get_bulk_action(bulk_action) if bulk_action else None

        if not action_func:
            tk.h.flash_error(tk._("The bulk action is not implemented"))
            return tk.redirect_to("file_manager.list")

        for file_id in file_ids:
            try:
                action_func(file_id)
            except tk.ValidationError as e:
                tk.h.flash_error(str(e))

        return tk.redirect_to("file_manager.list")

    def _get_bulk_action(self, value: str) -> Callable[[str], None] | None:
        return {
            "1": self._remove_file,
        }.get(value)

    def _remove_file(self, file_id: str) -> None:
        tk.get_action("files_file_delete")(
            {"ignore_auth": True},
            {"id": file_id},
        )


class FileManagerUploadView(MethodView):
    def post(self):
        file = tk.request.files.get("upload")

        if not file:
            tk.h.flash_error(tk._("Missing file object"))
            return tk.redirect_to("file_manager.list")

        try:
            tk.get_action("files_file_create")(
                {"ignore_auth": True},
                {
                    "name": file.filename,
                    "upload": file,
                },
            )
        except tk.ValidationError as e:
            tk.h.flash_error(str(e.error_summary))
            return tk.redirect_to("file_manager.list")
        except OSError as e:
            tk.h.flash_error(str(e))
            return tk.redirect_to("file_manager.list")

        tk.h.flash_success(tk._("File has been uploaded!"))
        return tk.redirect_to("file_manager.list")


class FileManagerDeleteView(MethodView):
    def post(self, file_id: str):
        try:
            tk.get_action("files_file_delete")({"ignore_auth": True}, {"id": file_id})
        except tk.ValidationError as e:
            tk.h.flash_error(str(e.error_summary))
            return tk.redirect_to("file_manager.list")
        except OSError as e:
            tk.h.flash_error(str(e))
            return tk.redirect_to("file_manager.list")

        tk.h.flash_success(tk._("File has been deleted!"))
        return tk.redirect_to("file_manager.list")


file_manager.add_url_rule("/manage", view_func=FileManagerView.as_view("list"))
file_manager.add_url_rule("/upload", view_func=FileManagerUploadView.as_view("upload"))
file_manager.add_url_rule(
    "/delete/<file_id>",
    view_func=FileManagerDeleteView.as_view("delete"),
)

blueprints = [file_manager]



import json
import os

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan.exceptions import CkanConfigurationException

from ckanext.files import (
    base,
    cli,
    config,
    exceptions,
    helpers,
    interfaces,
    storage,
    views,
)
from ckanext.files.logic import action, auth, validators

from ckanext.files import types  # isort: skip # noqa: F401


class FilesPlugin(p.SingletonPlugin):
    p.implements(p.IConfigurable)
    p.implements(p.IConfigurer, inherit=True)
    p.implements(p.IActions)
    p.implements(p.IAuthFunctions)
    p.implements(p.IValidators)
    p.implements(p.IBlueprint)
    p.implements(p.ITemplateHelpers)
    p.implements(interfaces.IFiles)

    if tk.check_ckan_version("2.9"):
        p.implements(p.IClick)

    if tk.check_ckan_version("2.10"):
        p.implements(p.IConfigDeclaration)

        def declare_config_options(self, declaration, key):
            # type: (types.Declaration, types.Key) -> None
            import yaml

            here = os.path.dirname(__file__)
            with open(os.path.join(here, "config_declaration.yaml"), "rb") as src:
                declaration.load_dict(yaml.safe_load(src))

            _register_adapters()
            for name, settings in config.storages().items():
                storage_key = key.from_string(config.STORAGE_PREFIX + name)

                if tk.check_ckan_version("2.10.3"):
                    available_adapters = json.dumps(
                        list(base.adapters),
                        separators=(",", ":"),
                    )

                    declaration.declare(
                        storage_key.type,
                        settings.get("type"),
                    ).append_validators(
                        "one_of({})".format(available_adapters),
                    ).set_description(
                        "Storage adapter used by the storage",
                    )

                adapter = base.adapters.get(settings.get("type"))
                if not adapter:
                    continue

                adapter.declare_config_options(
                    declaration,
                    storage_key,
                )

    # IFiles
    def files_get_storage_adapters(self):
        # type: () -> dict[str, types.Any]
        adapters = {}  # type: dict[str, types.Any]
        adapters = {
            "files:fs": storage.FileSystemStorage,
            "files:public_fs": storage.PublicFileSystemStorage,
            "files:redis": storage.RedisStorage,
        }

        if hasattr(storage, "GoogleCloudStorage"):
            adapters.update({"files:google_cloud_storage": storage.GoogleCloudStorage})

        return adapters

    # IConfigurable
    def configure(self, config_):
        # type: (types.Any) -> None

        # starting from CKAN v2.10, adapters are registered alongside with
        # config declaration, to enrich declarations with adapter-specific
        # options.
        if not tk.check_ckan_version("2.10"):
            _register_adapters()

        _initialize_storages()

    # IConfigurer
    def update_config(self, config_):
        # type: (types.Any) -> None
        tk.add_template_directory(config_, "templates")
        tk.add_resource("assets", "files")
        tk.add_public_directory(config_, "public")

    # IActions
    def get_actions(self):
        return action.get_actions()

    # IAuthFunctions
    def get_auth_functions(self):
        return auth.get_auth_functions()

    # IValidators
    def get_validators(self):
        return validators.get_validators()

    # IBlueprint
    def get_blueprint(self):
        return views.get_blueprints()

    # ITemplateHelpers
    def get_helpers(self):
        return helpers.get_helpers()

    # IClick
    def get_commands(self):
        return cli.get_commands()


def _register_adapters():
    """Register all storage types provided by extensions."""

    base.adapters.reset()
    for plugin in p.PluginImplementations(interfaces.IFiles):
        for name, adapter in plugin.files_get_storage_adapters().items():
            base.adapters.register(name, adapter)


def _initialize_storages():
    """Initialize all configured storages.

    Raise an exception if storage type is not registered.
    """

    base.storages.reset()
    for name, settings in config.storages().items():
        try:
            storage = base.storage_from_settings(name, settings)
        except exceptions.UnknownAdapterError as err:
            raise CkanConfigurationException(str(err))  # noqa: B904

        base.storages.register(name, storage)



"""Interfaces of the extension.
"""

import six

if six.PY3:
    from typing import Any  # isort: skip # noqa: F401


from ckan.plugins import Interface


class IFiles(Interface):
    """Extension point for ckanext-files."""

    def files_get_storage_adapters(self):
        # type: () -> dict[str, Any]
        """Return mapping of storage type to storage factory.

        Example:
        >>> def files_get_storage_adapters(self):
        >>>     return {
        >>>         "my_ext:dropbox": DropboxStorage,
        >>>     }

        """

        return {}



import copy

import six
import sqlalchemy as sa
from sqlalchemy import Column, DateTime, UnicodeText
from sqlalchemy.dialects.postgresql import JSONB

import ckan.plugins.toolkit as tk
from ckan.lib.dictization import table_dictize
from ckan.model.types import make_uuid

from .base import Base, now

from datetime import datetime  # isort: skip # noqa: F401


from ckanext.files import types  # isort: skip # noqa: F401

if six.PY3:
    from typing import Any, Literal  # isort: skip # noqa: F401


class File(Base):  # type: ignore
    __tablename__ = "files_file"
    id = Column(UnicodeText, primary_key=True, default=make_uuid)
    name = Column(UnicodeText, nullable=False)
    storage = Column(UnicodeText, nullable=False)

    ctime = Column(DateTime, nullable=False, default=now, server_default=sa.func.now())
    mtime = Column(DateTime)
    atime = Column(DateTime)

    storage_data = Column(JSONB, default=dict, server_default="{}")
    plugin_data = Column(JSONB, default=dict, server_default="{}")
    completed = Column(sa.Boolean, default=False, server_default="false")

    def __init__(self, **kwargs):
        # type: (**types.Any) -> None
        super(File, self).__init__(**kwargs)
        if not self.id:
            self.id = make_uuid()

    def dictize(self, context):
        # type: (Any) -> dict[str, Any]

        result = table_dictize(self, context)
        result["storage_data"] = copy.deepcopy(result["storage_data"])

        plugin_data = result.pop("plugin_data")
        if context.get("include_plugin_data"):
            result["plugin_data"] = copy.deepcopy(plugin_data)

        return result

    def touch(self, access=True, modification=True, moment=None):
        # type: (bool, bool, datetime | None) -> None
        if not moment:
            moment = now()

        if access:
            self.atime = moment

        if modification:
            self.mtime = moment

    def patch_data(self, patch, dict_path=None, prop="plugin_data"):
        # type: (dict[str, Any], list[str]|None, Literal["storage_data", "plugin_data"]) -> dict[str, Any]
        data = copy.deepcopy(getattr(self, prop))  # type: dict[str, Any]

        target = data  # type: dict[str, Any] | Any
        if dict_path:
            for part in dict_path:
                target = target.setdefault(part, {})
                if not isinstance(target, dict):
                    raise TypeError(part)
        target.update(patch)

        setattr(self, prop, data)
        return data

    @classmethod
    def by_location(cls, location, storage=None):
        # type: (str, str | None) -> sa.sql.Select
        selectable = cls if tk.check_ckan_version("2.9") else [cls]

        stmt = sa.select(selectable).where(
            cls.storage_data["filename"].astext == location,
        )
        if storage:
            stmt = stmt.where(cls.storage == storage)

        return stmt



import pytest
from faker import Faker  # pytest: skip # noqa: F401
from werkzeug.datastructures import FileStorage

import ckan.plugins.toolkit as tk

from ckanext.files.logic import validators


class TestIntoUpload:
    def test_file_storage(self):
        """FileStorage instances returned as-is."""
        upload = FileStorage()

        assert validators.files_into_upload(upload) is upload

    def test_str(self, faker):
        # type: (Faker) -> None
        """Strings converted into FileStorage."""
        string = faker.pystr()
        upload = validators.files_into_upload(string)

        assert isinstance(upload, FileStorage)
        assert upload.stream.read() == string.encode()

    def test_bytes(self, faker):
        # type: (Faker) -> None
        """Bytes converted into FileStorage."""
        binary = faker.binary(100)
        upload = validators.files_into_upload(binary)

        assert isinstance(upload, FileStorage)
        assert upload.stream.read() == binary

    def test_wrong_type(self):
        """Any unexpected value causes an exception."""
        with pytest.raises(tk.Invalid):
            validators.files_into_upload(123)


class TestParseFilesize:
    def test_int(self, faker):
        # type: (Faker) -> None
        """Numbers are not changed by validator."""

        value = faker.pyint()
        assert validators.files_parse_filesize(value) is value

    def test_str(self):
        """Strings are converted into a number."""
        assert validators.files_parse_filesize("10K") == 10000

    def test_wrong_value(self):
        """Any unexpected value causes an exception."""
        with pytest.raises(tk.Invalid):
            validators.files_parse_filesize("10P")



"""Add file last_access field

Revision ID: 2c5f1f90888c
Revises: cc1a832108c5
Create Date: 2024-01-27 12:47:35.568291

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2c5f1f90888c"
down_revision = "cc1a832108c5"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "files_file",
        sa.Column(
            "last_access",
            sa.DateTime(),
            nullable=False,
            server_default=sa.func.now(),
        ),
    )


def downgrade():
    op.drop_column("files_file", "last_access")



# type: ignore
from __future__ import print_function

import logging

import paste.script

from ckan import model
from ckan.lib.cli import CkanCommand

from ckanext.files.model.base import metadata

log = logging.getLogger(__name__)


def create_tables():
    """Create tables defined in model."""
    metadata.create_all(model.meta.engine)


class FilesCommand(CkanCommand):
    """
    ckanext-files management commands.

    Usage::
        paster --plugin=ckanext-files files  -c ckan.ini initdb
        paster --plugin=ckanext-files files  -c ckan.ini createdb
    """

    summary = __doc__.split("\n")[0]
    usage = __doc__

    parser = paste.script.command.Command.standard_parser(verbose=True)
    parser.add_option(
        "-c",
        "--config",
        dest="config",
        default="ckan.ini",
        help="Config file to use.",
    )

    def command(self):
        """Command itself."""
        self._load_config()

        if not len(self.args):
            print(self.usage)  # noqa

        elif self.args[0] == "initdb":
            self._init()
        elif self.args[0] == "createdb":
            self._create()

    def _init(self):
        self._create()
        log.info("DB tables are reinitialized")

    def _create(self):
        model.Session.rollback()

        create_tables()
        log.info("DB tables are setup")



import os

import pytest

from ckan.tests import factories
from ckan.tests.helpers import call_action

from ckanext.files.base import get_storage
from ckanext.files.exceptions import UnknownStorageError


@pytest.fixture()
def with_files_uploader(monkeypatch, tmpdir, ckan_config):
    monkeypatch.setitem(
        ckan_config,
        "ckanext.files.storage.resource.type",
        "files_uploader:resource",
    )
    monkeypatch.setitem(ckan_config, "ckanext.files.storage.resource.path", str(tmpdir))
    monkeypatch.setitem(
        ckan_config,
        "ckanext.files.storage.image.type",
        "files_uploader:image",
    )
    monkeypatch.setitem(ckan_config, "ckanext.files.storage.image.path", str(tmpdir))


class SharedTests:
    @pytest.mark.ckan_config("ckan.upload.group.mimetypes", "application/octet-stream")
    @pytest.mark.ckan_config("ckan.upload.group.types", "application")
    def test_group_upload(self, tmpdir, create_with_upload, faker):
        user = factories.User()
        group = create_with_upload(
            faker.binary(10),
            "image.png",
            context={"user": user["name"]},
            upload_field_name="image_upload",
            image_url="image.png",
            action="group_create",
            name=faker.word(),
        )
        filepath = os.path.join(
            tmpdir,
            "storage",
            "uploads",
            "group",
            group["image_url"],
        )
        assert os.path.exists(filepath)

    def test_resource_upload_goes_to_fs(self, tmpdir, create_with_upload):
        content = "hello world"
        package = factories.Dataset()
        resource = create_with_upload(content, "file.txt", package_id=package["id"])

        filepath = os.path.join(
            tmpdir,
            "resources",
            resource["id"][:3],
            resource["id"][3:6],
            resource["id"][6:],
        )
        assert os.path.exists(filepath)

        with open(filepath, "r") as src:
            assert src.read() == content

    def test_resource_download(self, app, create_with_upload):
        content = "hello world"
        package = factories.Dataset()
        resource = create_with_upload(content, "file.txt", package_id=package["id"])

        resp = app.get(
            "/dataset/{package_id}/resource/{id}/download".format(**resource),
        )
        assert resp.body == content

    def test_resource_deletion(self, create_with_upload, faker, tmpdir):
        """Files are kept in system when resource is deleted but removed when
        `clear_upload` field added to resource."""

        package = factories.Dataset()
        resource = create_with_upload(
            faker.binary(10),
            "file.txt",
            package_id=package["id"],
        )
        filepath = os.path.join(
            tmpdir,
            "resources",
            resource["id"][:3],
            resource["id"][3:6],
            resource["id"][6:],
        )
        assert os.path.exists(filepath)

        call_action("resource_patch", id=resource["id"], clear_upload=True)
        assert not os.path.exists(filepath)


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestNativeUploader(SharedTests):
    def test_resource_storage_is_not_registered(self):
        with pytest.raises(UnknownStorageError):
            get_storage("resource")


@pytest.mark.usefixtures("with_files_uploader", "with_plugins", "clean_db")
@pytest.mark.ckan_config("ckan.plugins", "files files_uploader")
class TestFilesUploader(SharedTests):
    def test_resource_storage_is_registered(self):
        assert get_storage("resource")



import base64
import hashlib
import json
import os
from io import BytesIO
from uuid import UUID

import pytest
import six
from six.moves.urllib.parse import quote_plus
from werkzeug.datastructures import FileStorage

import ckan.plugins.toolkit as tk

from ckanext.files import exceptions
from ckanext.files.storage import google_cloud as gc

from faker import Faker  # isort: skip # noqa: F401
from ckanext.files import types  # isort: skip # noqa: F401

TEST_BUCKET = "ld-bq-test"


@pytest.fixture()
def upload_progress():
    return {"size": 0, "uploaded": 0, "content": b"", "name": ""}


@pytest.fixture()
def storage():
    # type: () -> gc.GoogleCloudStorage
    return gc.GoogleCloudStorage(name="test", path="test", bucket=TEST_BUCKET)


@pytest.fixture()
def mocked_token(responses):
    responses.add(
        "POST",
        "https://oauth2.googleapis.com/token",
        json={
            "access_token": "ya29.c.c0AY_VpZgccnsCOsCGNFxy",
            "expires_in": 3599,
            "token_type": "Bearer",
        },
    )


def encode(content):
    # type: (bytes) -> str
    digest = hashlib.md5(content).digest()

    if six.PY3:
        return base64.encodebytes(digest).decode().strip()

    return base64.encodestring(digest).strip()


@pytest.fixture()
def mocked_session_url():
    return (
        "https://storage.googleapis.com"
        + "/upload/storage/v1/b/{}/o?".format(TEST_BUCKET)
        + "uploadType=resumable&upload_id=test"
    )


@pytest.fixture()
def mocked_upload_initialize(
    responses,
    mocked_session_url,
    mocked_token,
    upload_progress,
):
    # type: (types.Any, types.Any, types.Any, dict[str, types.Any]) -> None
    def post_callback(request):
        upload_progress["size"] = int(request.headers.get("x-upload-content-length", 0))
        upload_progress["name"] = json.loads(request.body)["name"]
        return (200, {"location": mocked_session_url}, "")

    responses.add_callback(
        "POST",
        "https://storage.googleapis.com/upload/storage/v1/b/{}/o?uploadType=resumable".format(
            TEST_BUCKET,
        ),
        callback=post_callback,
    )


@pytest.fixture()
def mocked_upload_update(responses, mocked_session_url, upload_progress):
    # type: (types.Any, types.Any, dict[str, types.Any]) -> None

    def put_callback(request):
        # type: (types.Any) -> types.Any

        content_range = request.headers.get("content-range")
        headers = {}
        if content_range:
            range, total = content_range.split()[-1].split("/")
            if range == "*":
                if upload_progress["uploaded"] == upload_progress["size"]:
                    status = 200
                else:
                    status = 308
                    if upload_progress["uploaded"]:
                        headers["range"] = "bytes=0-{}".format(
                            upload_progress["uploaded"],
                        )
            else:
                total = int(total)
                start, end = map(int, range.split("-"))
                upload_progress["uploaded"] = end + 1
                upload_progress["content"] = (
                    upload_progress["content"][:start] + request.body
                )
                if end + 1 < total:
                    status = 308
                    headers["range"] = "bytes=0-{}".format(end)
                else:
                    status = 200

        else:
            status = 200

        return (
            status,
            headers,
            (
                ""
                if status == 308
                else json.dumps(
                    {
                        "size": str(upload_progress["uploaded"]),
                        "md5Hash": encode(upload_progress["content"]),
                        "contentType": "application/octet-stream",
                        "name": upload_progress["name"],
                    },
                )
            ),
        )

    responses.add_callback(
        "PUT",
        mocked_session_url,
        callback=put_callback,
    )


@pytest.fixture()
def mocked_upload(mocked_upload_initialize, mocked_upload_update):
    # type: (types.Any, types.Any) -> None
    pass


class TestUploader:
    @pytest.mark.usefixtures("mocked_token", "mocked_upload")
    def test_result(self, storage, faker):
        # type: (gc.GoogleCloudStorage, Faker) -> None
        content = faker.binary(100)
        result = storage.upload("", FileStorage(BytesIO(content)), {})

        assert UUID(result["filename"])
        assert result["size"] == len(content)
        assert result["hash"] == gc.decode(encode(content))


@pytest.mark.usefixtures("with_plugins")
class TestMultipartUploader:
    def test_initialization_invalid(self, storage, faker):
        # type: (gc.GoogleCloudStorage, Faker) -> None
        with pytest.raises(tk.ValidationError):
            storage.initialize_multipart_upload(faker.file_name(), {})

    def test_initialization_large(self, storage, faker):
        # type: (gc.GoogleCloudStorage, Faker) -> None
        storage.settings["max_size"] = 5
        with pytest.raises(exceptions.LargeUploadError):
            storage.initialize_multipart_upload(faker.file_name(), {"size": 10})

    @pytest.mark.usefixtures("mocked_upload_initialize")
    def test_initialization(self, storage, faker):
        # type: (gc.GoogleCloudStorage, Faker) -> None
        content = b"hello world"
        data = storage.initialize_multipart_upload(
            faker.file_name(),
            {"size": len(content)},
        )
        assert data["size"] == len(content)
        assert data["uploaded"] == 0
        assert data["session_url"]

    @pytest.mark.usefixtures("mocked_upload_initialize")
    def test_update_invalid(self, storage, faker):
        # type: (gc.GoogleCloudStorage, Faker) -> None
        content = b"hello world"
        data = storage.initialize_multipart_upload(
            faker.file_name(),
            {"size": len(content)},
        )
        with pytest.raises(tk.ValidationError):
            storage.update_multipart_upload(data, {})

    @pytest.mark.usefixtures("mocked_upload")
    def test_update(self, storage, faker):
        # type: (gc.GoogleCloudStorage, Faker) -> None
        content = faker.binary(256 * 1024 * 2)
        data = storage.initialize_multipart_upload(
            faker.file_name(),
            {"size": len(content)},
        )

        with pytest.raises(tk.ValidationError):
            storage.update_multipart_upload(
                data,
                {"upload": FileStorage(BytesIO(content[:5]))},
            )

        data = storage.update_multipart_upload(
            data,
            {"upload": FileStorage(BytesIO(content[: 256 * 1024]))},
        )

        assert data["size"] == len(content)
        assert data["uploaded"] == 256 * 1024

        with pytest.raises(exceptions.UploadOutOfBoundError):
            storage.update_multipart_upload(
                data,
                {"upload": FileStorage(BytesIO(content))},
            )

        missing_size = data["size"] - data["uploaded"]
        data = storage.update_multipart_upload(
            data,
            {"upload": FileStorage(BytesIO(content[-missing_size:]))},
        )
        assert data["size"] == len(content)
        assert data["uploaded"] == len(content)

    @pytest.mark.usefixtures("mocked_upload")
    def test_complete(self, storage, faker):
        # type: (gc.GoogleCloudStorage, Faker) -> None
        content = b"hello world"
        data = storage.initialize_multipart_upload(
            faker.file_name(),
            {"size": len(content)},
        )

        with pytest.raises(tk.ValidationError):
            storage.complete_multipart_upload(data, {})

        data = storage.update_multipart_upload(
            data,
            {"upload": FileStorage(BytesIO(content))},
        )
        data = storage.complete_multipart_upload(data, {})
        assert data["size"] == len(content)
        assert data["hash"] == hashlib.md5(content).hexdigest()

    @pytest.mark.usefixtures("mocked_upload")
    def test_show(self, storage, faker):
        # type: (gc.GoogleCloudStorage, Faker) -> None
        content = b"hello world"

        data = storage.initialize_multipart_upload(
            faker.file_name(),
            {"size": len(content)},
        )
        assert storage.show_multipart_upload(data) == data

        data = storage.update_multipart_upload(
            data,
            {"upload": FileStorage(BytesIO(content))},
        )
        assert storage.show_multipart_upload(data) == data

        storage.complete_multipart_upload(data, {})
        assert storage.show_multipart_upload(data) == data


class TestManager:
    @pytest.mark.usefixtures("mocked_upload")
    def test_removal(self, storage, responses):
        # type: (gc.GoogleCloudStorage, types.Any) -> None
        result = storage.upload("", FileStorage(), {})
        name = os.path.join(storage.settings["path"], result["filename"])

        object_url = (
            "https://storage.googleapis.com/storage/v1/b/ld-bq-test/o/{}".format(
                quote_plus(name),
            )
        )
        responses.add("GET", object_url, status=200)
        responses.add("DELETE", object_url, status=200)

        assert storage.remove(result)

    @pytest.mark.usefixtures("mocked_upload")
    def test_removal_missing(self, storage, responses):
        # type: (gc.GoogleCloudStorage, types.Any) -> None
        result = storage.upload("", FileStorage(), {})
        name = os.path.join(storage.settings["path"], result["filename"])

        object_url = (
            "https://storage.googleapis.com/storage/v1/b/ld-bq-test/o/{}".format(
                quote_plus(name),
            )
        )
        responses.add("GET", object_url, status=200)
        responses.add("GET", object_url, status=404)

        responses.add("DELETE", object_url, status=200)

        assert storage.remove(result)
        assert not storage.remove(result)


class TestStorage:
    def test_missing_path(self, tmp_path):
        # type: (str) -> None
        with pytest.raises(exceptions.InvalidStorageConfigurationError):
            gc.GoogleCloudStorage(bucket=TEST_BUCKET, credentials_file="/not-real")



import logging
from functools import partial

from flask import Blueprint
from flask.views import MethodView

import ckan.plugins.toolkit as tk
from ckan.lib.helpers import Page

from ckanext.files import types  # isort: skip # noqa: F401

log = logging.getLogger(__name__)
bp = Blueprint("files", __name__)


def not_found_handler(error):
    # type: (tk.ObjectNotFound) -> tuple[str, int]
    """Generic handler for ObjectNotFound exception"""
    return (
        tk.render(
            "error_document_template.html",
            {
                "code": 404,
                "content": "Object not found: {}".format(error.message),
                "name": "Not found",
            },
        ),
        404,
    )


def not_authorized_handler(error):
    # type: (tk.NotAuthorized) -> tuple[str, int]
    """Generic handler for NotAuthorized exception"""
    return (
        tk.render(
            "error_document_template.html",
            {
                "code": 403,
                "content": error.message or "Not authorized to view this page",
                "name": "Not authorized",
            },
        ),
        403,
    )


bp.register_error_handler(tk.ObjectNotFound, not_found_handler)
bp.register_error_handler(tk.NotAuthorized, not_authorized_handler)


def get_blueprints():
    return [bp]


def _pager_url(*args, **kwargs):
    # type: (*types.Any, **types.Any) -> str
    """Generic URL builder for `url` parameter of ckan.lib.pagination.Page.

    It generates pagination link keeping all the parameters from URL and query
    string.
    """
    params = {k: v for k, v in tk.request.args.items() if k != "page"}
    params.update({k: v for k, v in (tk.request.view_args or {}).items()})
    params.update(kwargs)
    return tk.h.pager_url(*args, **params)


@bp.route("/user/<user_id>/files")
@bp.route("/user/<user_id>/files/storage/<storage>")
def user(user_id, storage=None):
    # type: (str, str | None) -> str
    user_dict = tk.get_action("user_show")(
        {},
        {"id": user_id, "include_num_followers": True},
    )

    rows = 10
    params = tk.request.args if tk.check_ckan_version("2.10") else tk.request.params  # type: ignore
    page = tk.h.get_page_number(params)
    start = rows * page - rows

    search_dict = {
        "rows": rows,
        "start": start,
        "user": user_id,
        "sort": params.get("sort", "ctime"),
        "reverse": params.get("reverse", True),
    }  # type: dict[str, types.Any]

    if storage:
        search_dict["storage"] = storage

    try:
        files = tk.get_action("files_file_search_by_user")({}, search_dict)
    except tk.ValidationError as err:
        for k, v in err.error_summary.items():
            tk.h.flash_error("{}: {}".format(k, v))

        files = {"count": 0, "results": []}  # type: dict[str, types.Any]

    pager = Page(
        [],
        items_per_page=rows,
        page=page,
        item_count=files["count"],
        url=partial(_pager_url),
    )

    tpl_names = ["files/user/index.html"]
    tpl_data = {
        "user_dict": user_dict,
        "files": files,
        "pager": pager,
    }  # type: dict[str, types.Any]

    if storage:
        tpl_data["storage"] = storage
        tpl_names.insert(0, "files/user/index.{}.html".format(storage))

    return tk.render(tpl_names, tpl_data)  # type: ignore


class DeleteFile(MethodView):
    def post(self, user_id, file_id):
        # type: (str, str) -> types.Any

        try:
            tk.get_action("files_file_delete")({}, {"id": file_id})
        except tk.NotAuthorized as err:
            tk.h.flash_error(err)
            return self.get(user_id, file_id)

        came_from = tk.h.get_request_param("came_from")
        if came_from:
            return tk.redirect_to(came_from)

        return tk.redirect_to("files.user", user_id=user_id)

    def get(self, user_id, file_id):
        # type: (str, str) -> types.Any

        tk.check_access("files_file_delete", {}, {"id": file_id})
        info = tk.get_action("files_file_show")({}, {"id": file_id})
        user_dict = tk.get_action("user_show")(
            {},
            {"id": user_id, "include_num_followers": True},
        )

        return tk.render(
            "files/user/delete.html",
            {
                "file": info,
                "user_dict": user_dict,
            },
        )


bp.add_url_rule(
    "/user/<user_id>/files/delete/<file_id>",
    view_func=DeleteFile.as_view("delete_file"),
)



"""merge upload and file models

Revision ID: 64ca007bf3eb
Revises: 2fbd30a1b364
Create Date: 2024-03-07 21:01:21.133717

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import JSONB

# revision identifiers, used by Alembic.
revision = "64ca007bf3eb"
down_revision = "2fbd30a1b364"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table("files_upload")
    op.add_column(
        "files_file",
        sa.Column("completed", sa.Boolean, server_default="false"),
    )


def downgrade():
    op.create_table(
        "files_upload",
        sa.Column("id", sa.Text, primary_key=True),
        sa.Column("name", sa.Text, nullable=False),
        sa.Column("storage", sa.Text, nullable=False),
        sa.Column(
            "initialized_at",
            sa.DateTime,
            nullable=False,
            server_default=sa.func.now(),
        ),
        sa.Column("upload_data", JSONB, server_default="{}"),
    )
    op.drop_column("files_file", "completed")



"""create upload table

Revision ID: 3c69eb68cecd
Revises: 5851e09b7ca3
Create Date: 2024-03-03 21:53:32.955167

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import JSONB

# revision identifiers, used by Alembic.
revision = "3c69eb68cecd"
down_revision = "5851e09b7ca3"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "files_upload",
        sa.Column("id", sa.Text, primary_key=True),
        sa.Column("name", sa.Text, nullable=False),
        sa.Column("storage", sa.Text, nullable=False),
        sa.Column(
            "initialized_at",
            sa.DateTime,
            nullable=False,
            server_default=sa.func.now(),
        ),
        sa.Column("upload_data", JSONB, server_default="{}"),
    )


def downgrade():
    op.drop_table("files_upload")



import os

import ckan.plugins.toolkit as tk

from ckanext.files.utils import make_collector

_helpers, helper = make_collector()

HERE = os.path.dirname(__file__)


def get_helpers():
    return dict(_helpers)


@helper
def files_humanize_content_type(content_type):
    # type: (str) -> str
    if not content_type:
        content_type = "application/octet-stream"

    # for name, types in tk.h.resource_formats().items():
    #     if name == content_type or content_type in types:
    #         return types[2]

    main, _sub = content_type.split("/")
    return main.capitalize()


@helper
def files_content_type_icon(content_type, theme, extension):
    # type: (str, str, str) -> str | None
    if not content_type:
        content_type = "application/octet-stream"

    url = "ckanext-files/mimetype_icons/{}/{}.{}".format(
        theme,
        content_type.replace("/", "-"),
        extension,
    )

    if os.path.exists(os.path.join(HERE, "public", url)):
        return tk.h.url_for_static(url)

    return None



import hashlib
from io import BytesIO
from uuid import UUID

import pytest
from werkzeug.datastructures import FileStorage

from ckanext.files import exceptions
from ckanext.files.storage import redis

from faker import Faker  # isort: skip # noqa: F401


@pytest.fixture()
def storage(clean_redis):
    # type: (object) -> redis.RedisStorage
    return redis.RedisStorage(name="test")


class TestUploader:
    def test_key(self, storage):
        # type: (redis.RedisStorage) -> None
        result = storage.upload("", FileStorage(), {})

        assert UUID(result["filename"])
        assert result["size"] == 0

        key = storage.settings["prefix"] + result["filename"]
        assert storage.redis.get(key) == b""

    def test_content(self, storage, faker):
        # type: (redis.RedisStorage, Faker) -> None
        content = faker.binary(100)
        result = storage.upload("", FileStorage(BytesIO(content)), {})

        assert result["size"] == 100

        key = storage.settings["prefix"] + result["filename"]
        assert storage.redis.get(key) == content

    def test_hash(self, storage, faker):
        # type: (redis.RedisStorage, Faker) -> None
        result = storage.upload("", FileStorage(), {})
        assert result["hash"] == hashlib.md5().hexdigest()

        content = faker.binary(100)
        result = storage.upload("", FileStorage(BytesIO(content)), {})
        assert result["hash"] == hashlib.md5(content).hexdigest()

    def test_copy(self, storage, faker):
        # type: (redis.RedisStorage, Faker) -> None
        content = faker.binary(100)
        original = storage.upload("", FileStorage(BytesIO(content)), {})
        copy = storage.copy(original, storage, faker.file_name(), {})

        assert copy["filename"] != original["filename"]
        assert storage.content(copy) == storage.content(original)

    def test_move(self, storage, faker):
        # type: (redis.RedisStorage, Faker) -> None
        content = faker.binary(100)
        original = storage.upload("", FileStorage(BytesIO(content)), {})
        copy = storage.move(original, storage, faker.file_name(), {})

        with pytest.raises(exceptions.MissingFileError):
            storage.content(original)

        assert storage.content(copy) == content


class TestManager:
    def test_removal(self, storage):
        # type: (redis.RedisStorage) -> None
        result = storage.upload("", FileStorage(), {})
        key = storage.settings["prefix"] + result["filename"]

        assert storage.redis.exists(key)

        storage.remove(result)
        assert not storage.redis.exists(key)

    def test_exists(self, storage):
        # type: (redis.RedisStorage) -> None
        result = storage.upload("", FileStorage(), {})

        assert storage.exists(result)
        storage.remove(result)
        assert not storage.exists(result)


class TestReader:
    def test_stream(self, storage, faker):
        # type: (redis.RedisStorage, Faker) -> None
        data = faker.binary(100)
        result = storage.upload("", FileStorage(BytesIO(data)), {})

        stream = storage.stream(result)

        assert stream.read() == data

    def test_content(self, storage, faker):
        # type: (redis.RedisStorage, Faker) -> None
        data = faker.binary(100)
        result = storage.upload("", FileStorage(BytesIO(data)), {})

        content = storage.content(result)

        assert content == data

    def test_missing(self, storage, faker):
        # type: (redis.RedisStorage, Faker) -> None
        result = storage.upload("", FileStorage(), {})
        result["filename"] += faker.uuid4()

        with pytest.raises(exceptions.MissingFileError):
            storage.stream(result)

        with pytest.raises(exceptions.MissingFileError):
            storage.content(result)



import six

import ckan.plugins.toolkit as tk

from ckanext.files.utils import make_collector

from ckanext.files import types, utils  # isort: skip # noqa: F401

if six.PY3:
    from typing import Any  # isort: skip # noqa: F401

_validators, validator = make_collector()


def get_validators():
    return dict(_validators)


@validator
def files_into_upload(value):
    # type: (Any) -> types.Upload
    """Convert value into werkzeug.FileStorage object"""
    try:
        return utils.make_upload(value)

    except TypeError as err:
        msg = "Unsupported source type: {}".format(err)
        raise tk.Invalid(msg)  # noqa: B904

    except ValueError as err:
        msg = "Wrong file: {}".format(err)
        raise tk.Invalid(msg)  # noqa: B904


@validator
def files_parse_filesize(value):
    # type: (Any) -> int
    """Convert human-readable filesize into an integer."""

    if isinstance(value, int):
        return value

    try:
        return utils.parse_filesize(value)
    except ValueError:
        raise tk.Invalid("Wrong filesize string: {}".format(value))  # noqa: B904



from ckan.logic.schema import validator_args

from ckanext.files import config

from ckanext.files import types  # isort: skip # noqa: F401


@validator_args
def file_create(ignore_empty, unicode_safe, default, files_into_upload, not_missing):
    # type: (types.Any, types.Any, types.Any, types.Any, types.Any) -> types.Any

    # name is checked inside action, using "upload" as source if empty
    return {
        "name": [ignore_empty, unicode_safe],
        "storage": [default(config.default_storage()), unicode_safe],
        "upload": [not_missing, files_into_upload],
    }


@validator_args
def _base_file_search(
    unicode_safe,
    default,
    int_validator,
    boolean_validator,
    ignore_empty,
):
    # type: (types.Any, types.Any, types.Any, types.Any, types.Any) -> types.Any

    return {
        "start": [default(0), int_validator],
        "rows": [default(10), int_validator],
        "sort": [default("name"), unicode_safe],
        "reverse": [boolean_validator],
        "storage": [ignore_empty, unicode_safe],
        "storage_data": [ignore_empty],
        "plugin_data": [ignore_empty],
    }


@validator_args
def file_search_by_user(ignore_missing, unicode_safe, default, ignore_not_sysadmin):
    # type: (types.Any, types.Any, types.Any, types.Any) -> types.Any
    schema = _base_file_search()
    schema["user"] = [ignore_missing, ignore_not_sysadmin, unicode_safe]
    return schema


@validator_args
def file_delete(not_empty, unicode_safe):
    # type: (types.Any, types.Any) -> types.Any
    return {
        "id": [not_empty, unicode_safe],
    }


@validator_args
def file_show(not_empty, unicode_safe):
    # type: (types.Any, types.Any) -> types.Any
    return {
        "id": [not_empty, unicode_safe],
    }


@validator_args
def file_rename(not_empty, unicode_safe):
    # type: (types.Any, types.Any) -> types.Any
    return {
        "id": [not_empty, unicode_safe],
        "name": [not_empty, unicode_safe],
    }


@validator_args
def upload_initialize(ignore_empty, unicode_safe, default, int_validator, not_missing):
    # type: (types.Any, types.Any, types.Any, types.Any, types.Any) -> types.Any

    # name is checked inside action, using "upload" as source if empty
    return {
        "name": [ignore_empty, unicode_safe],
        "storage": [default(config.default_storage()), unicode_safe],
    }


@validator_args
def upload_show(not_empty, unicode_safe):
    # type: (types.Any, types.Any) -> types.Any
    return {
        "id": [not_empty, unicode_safe],
    }


@validator_args
def upload_update(not_empty, unicode_safe):
    # type: (types.Any, types.Any) -> types.Any
    return {
        "id": [not_empty, unicode_safe],
    }


@validator_args
def upload_complete(not_empty, unicode_safe):
    # type: (types.Any, types.Any) -> types.Any
    return {
        "id": [not_empty, unicode_safe],
    }



import base64
import os
import re

import requests
import six
from google.api_core.exceptions import Forbidden
from google.cloud.storage import Client
from google.oauth2.service_account import Credentials

import ckan.plugins.toolkit as tk

from ckanext.files import exceptions, types, utils
from ckanext.files.base import Capability, Manager, Storage, Uploader

GCAdditionalData = types.TypedDict("GCAdditionalData", {})


class GCStorageData(GCAdditionalData, types.MinimalStorageData):
    pass


RE_RANGE = re.compile(r"bytes=(?P<first_byte>\d+)-(?P<last_byte>\d+)")


def decode(value):
    # type: (str) -> str
    if six.PY3:
        return base64.decodebytes(value.encode()).hex()

    return base64.decodestring(value.encode()).encode("hex")  # type: ignore


class GoogleCloudUploader(Uploader):
    storage = None  # type: GoogleCloudStorage # pyright: ignore

    required_options = ["bucket"]
    capabilities = utils.combine_capabilities(
        Capability.CREATE,
        Capability.MULTIPART_UPLOAD,
    )

    def upload(self, name, upload, extras):
        # type: (str, types.Upload, dict[str, types.Any]) -> GCStorageData
        filename = self.storage.compute_name(name, extras, upload)
        filepath = os.path.join(self.storage.settings["path"], filename)

        client = self.storage.client
        blob = client.bucket(self.storage.settings["bucket"]).blob(filepath)

        blob.upload_from_file(upload.stream)
        filehash = decode(blob.md5_hash)
        return {
            "filename": filename,
            "content_type": upload.content_type,
            "hash": filehash,
            "size": blob.size or upload.content_length,
        }

    def initialize_multipart_upload(self, name, extras):
        # type: (str, dict[str, types.Any]) -> dict[str, types.Any]

        schema = {
            "size": [
                tk.get_validator("not_missing"),
                tk.get_validator("int_validator"),
            ],
            "__extras": [tk.get_validator("ignore")],
        }
        data, errors = tk.navl_validate(extras, schema)

        if errors:
            raise tk.ValidationError(errors)

        filename = self.storage.compute_name(name, extras)
        filepath = os.path.join(self.storage.settings["path"], filename)

        client = self.storage.client
        blob = client.bucket(self.storage.settings["bucket"]).blob(filepath)

        max_size = self.storage.max_size
        if max_size and data["size"] > max_size:
            raise exceptions.LargeUploadError(data["size"], max_size)

        url = blob.create_resumable_upload_session(
            size=data["size"],
            origin=self.storage.settings["resumable_origin"],
        )  # type: types.Any

        if not url:
            raise exceptions.UploadError("Cannot initialize session URL")

        return {
            "session_url": url,
            "size": data["size"],
            "uploaded": 0,
            "filename": filename,
        }

    def update_multipart_upload(self, upload_data, extras):
        # type: (dict[str, types.Any], dict[str, types.Any]) -> dict[str, types.Any]
        schema = {
            "upload": [
                tk.get_validator("ignore_missing"),
                tk.get_validator("files_into_upload"),
            ],
            "position": [
                tk.get_validator("ignore_missing"),
                tk.get_validator("int_validator"),
            ],
            "uploaded": [
                tk.get_validator("ignore_missing"),
                tk.get_validator("int_validator"),
            ],
            "__extras": [tk.get_validator("ignore")],
        }
        data, errors = tk.navl_validate(extras, schema)

        if errors:
            raise tk.ValidationError(errors)

        if "upload" in data:
            upload = data["upload"]  # type: types.Upload

            first_byte = data.get("position", upload_data["uploaded"])
            last_byte = first_byte + upload.content_length - 1
            size = upload_data["size"]

            if last_byte >= size:
                raise exceptions.UploadOutOfBoundError(last_byte, size)

            if upload.content_length < 256 * 1024 and last_byte < size - 1:
                raise tk.ValidationError(
                    {"upload": ["Only the final part can be smaller than 256KiB"]},
                )

            resp = requests.put(
                upload_data["session_url"],
                data=upload.stream.read(),
                headers={
                    "content-range": "bytes {}-{}/{}".format(
                        first_byte,
                        last_byte,
                        size,
                    ),
                },
            )

            if not resp.ok:
                raise tk.ValidationError({"upload": [resp.text]})

            if "range" not in resp.headers:
                upload_data["uploaded"] = upload_data["size"]
                upload_data["result"] = resp.json()
                return upload_data

            range_match = RE_RANGE.match(resp.headers["range"])
            if not range_match:
                raise tk.ValidationError(
                    {"upload": ["Invalid response from Google Cloud"]},
                )
            upload_data["uploaded"] = tk.asint(range_match.group("last_byte")) + 1

        elif "uploaded" in data:
            upload_data["uploaded"] = data["uploaded"]

        else:
            raise tk.ValidationError(
                {"upload": ["Either upload or uploaded must be specified"]},
            )

        return upload_data

    def show_multipart_upload(self, upload_data):
        # type: (dict[str, types.Any]) -> dict[str, types.Any]

        resp = requests.put(
            upload_data["session_url"],
            headers={
                "content-range": "bytes */{}".format(upload_data["size"]),
                "content-length": "0",
            },
        )
        if not resp.ok:
            raise tk.ValidationError({"id": [resp.text]})

        if resp.status_code == 308:
            if "range" in resp.headers:
                range_match = RE_RANGE.match(resp.headers["range"])
                if not range_match:
                    raise tk.ValidationError(
                        {
                            "id": [
                                "Invalid response from Google Cloud:"
                                + " missing range header",
                            ],
                        },
                    )
                upload_data["uploaded"] = tk.asint(range_match.group("last_byte")) + 1
            else:
                upload_data["uploaded"] = 0
        elif resp.status_code in [200, 201]:
            upload_data["uploaded"] = upload_data["size"]
            upload_data["result"] = resp.json()

        else:
            raise tk.ValidationError(
                {
                    "id": [
                        "Invalid response from Google Cloud:"
                        + " unexpected status {}".format(
                            resp.status_code,
                        ),
                    ],
                },
            )

        return upload_data

    def complete_multipart_upload(self, upload_data, extras):
        # type: (dict[str, types.Any], dict[str, types.Any]) -> GCStorageData
        upload_data = self.show_multipart_upload(upload_data)
        if upload_data["uploaded"] != upload_data["size"]:
            raise tk.ValidationError(
                {
                    "size": [
                        "Actual filesize {} does not match expected {}".format(
                            upload_data["uploaded"],
                            upload_data["size"],
                        ),
                    ],
                },
            )

        filehash = decode(upload_data["result"]["md5Hash"])

        return {
            "filename": os.path.relpath(
                upload_data["result"]["name"],
                self.storage.settings["path"],
            ),
            "hash": filehash,
            "content_type": upload_data["result"]["contentType"],
            "size": upload_data["size"],
        }


class GoogleCloudManager(Manager):
    storage = None  # type: GoogleCloudStorage # pyright: ignore
    required_options = ["bucket"]
    capabilities = utils.combine_capabilities(Capability.REMOVE)

    def remove(self, data):
        # type: (types.MinimalStorageData) -> bool

        filepath = os.path.join(str(self.storage.settings["path"]), data["filename"])
        client = self.storage.client  # type: Client
        blob = client.bucket(self.storage.settings["bucket"]).blob(filepath)

        try:
            exists = blob.exists()
        except Forbidden as err:
            raise exceptions.PermissionError(  # noqa: B904
                type(self),
                "exists",
                str(err),
            )

        if exists:
            try:
                blob.delete()
            except Forbidden as err:
                raise exceptions.PermissionError(  # noqa: B904
                    type(self),
                    "remove",
                    str(err),
                )
            return True
        return False


class GoogleCloudStorage(Storage):
    def __init__(self, **settings):
        # type: (**types.Any) -> None
        settings["path"] = settings.setdefault("path", "").lstrip("/")
        settings.setdefault("resumable_origin", tk.config["ckan.site_url"])

        super(GoogleCloudStorage, self).__init__(**settings)

        credentials = None
        credentials_file = settings.get("credentials_file", None)
        if credentials_file:
            try:
                credentials = Credentials.from_service_account_file(credentials_file)
            except IOError:
                raise exceptions.InvalidStorageConfigurationError(  # noqa: B904
                    type(self),
                    "file `{}` does not exist".format(credentials_file),
                )

        self.client = Client(credentials=credentials)

    def make_uploader(self):
        return GoogleCloudUploader(self)

    def make_manager(self):
        return GoogleCloudManager(self)

    @classmethod
    def declare_config_options(cls, declaration, key):
        # type: (types.Declaration, types.Key) -> None
        super().declare_config_options(declaration, key)
        declaration.declare(key.path, "").set_description(
            "Path to the folder where uploaded data will be stored.",
        )
        declaration.declare(key.bucket).required().set_description(
            "Name of the GCS bucket where uploaded data will be stored.",
        )
        declaration.declare(key.credentials_file).set_description(
            "Path to the credentials file used for authentication by GCS client."
            + "\nIf empty, uses value of GOOGLE_APPLICATION_CREDENTIALS envvar.",
        )
        declaration.declare(
            key.resumable_origin,
            tk.config["ckan.site_url"],
        ).set_description(
            "Value of the Origin header set for resumable upload."
            + "\nIn most cases, keep it the same as CKAN base URL.",
        )
